#!/usr/bin/env python3
#
# Copyright © 2020 Google, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from mako.template import Template
from isa import ISA
import argparse
import os
import sys

class FieldDecode(object):
    def __init__(self, name, map_expr):
        self.name = name
        self.map_expr = map_expr

    def get_c_name(self):
        return self.name.lower().replace('-', '_')

# State and helpers used by the template:
class State(object):
    def __init__(self, isa):
        self.isa = isa

    def case_name(self, bitset, name):
        return bitset.encode.case_prefix + name.upper().replace('.', '_').replace('-', '_').replace('#', '')

    # Return a list of all <map> entries for a leaf bitset, with the child
    # bitset overriding the parent bitset's entries. Because we can't resolve
    # which <map>s are used until we resolve which overload is used, we
    # generate code for encoding all of these and then at runtime select which
    # one to call based on the display.
    def decode_fields(self, bitset):
        if bitset.get_root().decode is None:
            return

        seen_fields = set()
        if bitset.encode is not None:
            for name, expr in bitset.encode.maps.items():
                seen_fields.add(name)
                yield FieldDecode(name, expr)

        if bitset.extends is not None:
            for field in self.decode_fields(self.isa.bitsets[bitset.extends]):
                if field.name not in seen_fields:
                    yield field

    # A limited resolver for field type which doesn't properly account for
    # overrides.  In particular, if a field is defined differently in multiple
    # different cases, this just blindly picks the last one.
    #
    # TODO to do this properly, I don't think there is an alternative than
    # to emit code which evaluates the case.expr
    def resolve_simple_field(self, bitset, name):
        field = None
        for case in bitset.cases:
            if name in case.fields:
                field = case.fields[name]
        if field is not None:
            return field
        if bitset.extends is not None:
            return self.resolve_simple_field(bitset.isa.bitsets[bitset.extends], name)
        return None

template = """\
/* Copyright (C) 2020 Google, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#include "${header}"

#include <stdint.h>
#include <util/bitset.h>

#define BITMASK_WORDS BITSET_WORDS(${isa.bitsize})

typedef struct {
    BITSET_WORD bitset[BITMASK_WORDS];
} bitmask_t;


#define BITSET_FORMAT ${isa.format()}
#define BITSET_VALUE(v) ${isa.value()}

static inline void
next_instruction(bitmask_t *instr, BITSET_WORD *start)
{
    %for i in range(0, int(isa.bitsize / 32)):
    instr->bitset[${i}] = *(start + ${i});
    %endfor
}

static inline uint64_t
bitmask_to_uint64_t(bitmask_t mask)
{
%   if isa.bitsize <= 32:
    return mask.bitset[0];
%   else:
    return ((uint64_t)mask.bitset[1] << 32) | mask.bitset[0];
%   endif
}

static inline bitmask_t
uint64_t_to_bitmask(uint64_t val)
{
    bitmask_t mask = {
        .bitset[0] = val & 0xffffffff,
%   if isa.bitsize > 32:
        .bitset[1] = (val >> 32) & 0xffffffff,
%   endif
    };

    return mask;
}

#include "isaspec_decode_decl.h"

static uint64_t
isa_decode_field(struct decode_scope *scope, const char *field_name);

/*
 * enum tables, these don't have any link back to other tables so just
 * dump them up front before the bitset tables
 */

%for name, enum in isa.enums.items():
static const struct isa_enum ${enum.get_c_name()} = {
    .num_values = ${len(enum.values)},
    .values = {
%   for val, display in enum.values.items():
        { .val = ${val}, .display = "${display}" },
%   endfor
    },
};
%endfor

/*
 * generated expression functions, can be linked from bitset tables, so
 * also dump them up front
 */

%for name, expr in isa.expressions.items():
static uint64_t
${expr.get_c_name()}(struct decode_scope *scope)
{
%   for fieldname in sorted(expr.fieldnames):
    int64_t ${fieldname} = isa_decode_field(scope, "${fieldname}");
%   endfor
    return ${expr.expr};
}
%endfor

/* forward-declarations of bitset decode functions */
%for name, bitset in isa.all_bitsets():
%   for df in s.decode_fields(bitset):
static void decode_${bitset.get_c_name()}_gen_${bitset.gen_min}_${df.get_c_name()}(void *out, struct decode_scope *scope, uint64_t val);
%   endfor
static const struct isa_field_decode decode_${bitset.get_c_name()}_gen_${bitset.gen_min}_fields[] = {
%   for df in s.decode_fields(bitset):
    {
        .name = "${df.name}",
        .decode = decode_${bitset.get_c_name()}_gen_${bitset.gen_min}_${df.get_c_name()},
    },
%   endfor
};
static void decode_${bitset.get_c_name()}_gen_${bitset.gen_min}(void *out, struct decode_scope *scope);
%endfor

/*
 * Forward-declarations (so we don't have to figure out which order to
 * emit various tables when they have pointers to each other)
 */

%for name, bitset in isa.all_bitsets():
static const struct isa_bitset bitset_${bitset.get_c_name()}_gen_${bitset.gen_min};
%endfor

%for root_name, root in isa.roots.items():
static const struct isa_bitset *${root.get_c_name()}[];
%endfor

/*
 * bitset tables:
 */

%for name, bitset in isa.all_bitsets():
%   for case in bitset.cases:
%      for field_name, field in case.fields.items():
%         if field.get_c_typename() == 'TYPE_BITSET':
%            if len(field.params) > 0:
static const struct isa_field_params ${case.get_c_name()}_gen_${bitset.gen_min}_${field.get_c_name()} = {
       .num_params = ${len(field.params)},
       .params = {
%               for param in field.params:
           { .name= "${param[0]}",  .as = "${param[1]}" },
%               endfor

       },
};
%            endif
%         endif
%      endfor
static const struct isa_case ${case.get_c_name()}_gen_${bitset.gen_min} = {
%   if case.expr is not None:
       .expr     = &${isa.expressions[case.expr].get_c_name()},
%   endif
%   if case.display is not None:
       .display  = "${case.display}",
%   endif
       .num_fields = ${len(case.fields)},
       .fields   = {
%   for field_name, field in case.fields.items():
          { .name = "${field_name}", .low = ${field.low}, .high = ${field.high},
%      if field.expr is not None:
            .expr = &${isa.expressions[field.expr].get_c_name()},
%      endif
%      if field.display is not None:
            .display = "${field.display}",
%      endif
            .type = ${field.get_c_typename()},
%      if field.get_c_typename() == 'TYPE_BITSET':
            .bitsets = ${isa.roots[field.type].get_c_name()},
%         if len(field.params) > 0:
            .params = &${case.get_c_name()}_gen_${bitset.gen_min}_${field.get_c_name()},
%         endif
%      endif
%      if field.get_c_typename() == 'TYPE_ENUM':
            .enums = &${isa.enums[field.type].get_c_name()},
%      endif
%      if field.get_c_typename() == 'TYPE_ASSERT':
            .val.bitset = { ${', '.join(isa.split_bits(field.val, 32))} },
%      endif
%      if field.get_c_typename() == 'TYPE_BRANCH' or field.get_c_typename() == 'TYPE_ABSBRANCH':
            .call = ${str(field.call).lower()},
%      endif
          },
%   endfor
       },
};
%   endfor
static const struct isa_bitset bitset_${bitset.get_c_name()}_gen_${bitset.gen_min} = {
<% pattern = bitset.get_pattern() %>
%   if bitset.extends is not None:
       .parent   = &bitset_${isa.bitsets[bitset.extends].get_c_name()}_gen_${isa.bitsets[bitset.extends].gen_min},
%   endif
       .name     = "${bitset.display_name}",
       .gen      = {
           .min  = ${bitset.get_gen_min()},
           .max  = ${bitset.get_gen_max()},
       },
       .match.bitset    = { ${', '.join(isa.split_bits(pattern.match, 32))} },
       .dontcare.bitset = { ${', '.join(isa.split_bits(pattern.dontcare, 32))} },
       .mask.bitset     = { ${', '.join(isa.split_bits(pattern.mask, 32))} },
       .decode = decode_${bitset.get_c_name()}_gen_${bitset.gen_min},
       .num_decode_fields = ARRAY_SIZE(decode_${bitset.get_c_name()}_gen_${bitset.gen_min}_fields),
       .decode_fields = decode_${bitset.get_c_name()}_gen_${bitset.gen_min}_fields,
       .num_cases = ${len(bitset.cases)},
       .cases    = {
%   for case in bitset.cases:
            &${case.get_c_name()}_gen_${bitset.gen_min},
%   endfor
       },
};
%endfor

/*
 * bitset hierarchy root tables (where decoding starts from):
 */

%for root_name, root in isa.roots.items():
static const struct isa_bitset *${root.get_c_name()}[] = {
%   for leaf_name, leafs in isa.leafs.items():
%      for leaf in leafs:
%         if leaf.get_root() == root:
             &bitset_${leaf.get_c_name()}_gen_${leaf.gen_min},
%         endif
%      endfor
%   endfor
    (void *)0
};
%endfor

#include "isaspec_decode_impl.c"

%for name, bitset in isa.all_bitsets():
%   for df in s.decode_fields(bitset):
<%  field = s.resolve_simple_field(bitset, df.name) %>
static void decode_${bitset.get_c_name()}_gen_${bitset.gen_min}_${df.get_c_name()}(void *out, struct decode_scope *scope, uint64_t val)
{
%       if bitset.get_root().decode is not None and field is not None:
    ${bitset.get_root().encode.type} src = *(${bitset.get_root().encode.type} *)out;
%           if field.get_c_typename() == 'TYPE_BITSET':
    isa_decode_bitset(&${df.map_expr}, ${isa.roots[field.type].get_c_name()}, scope, uint64_t_to_bitmask(val));
%           elif field.get_c_typename() in ['TYPE_BRANCH', 'TYPE_INT', 'TYPE_OFFSET']:
    ${df.map_expr} = util_sign_extend(val, ${field.get_size()});
%           else:
    ${df.map_expr} = val;
%           endif
    *(${bitset.get_root().encode.type} *)out = src;
%       endif
}

%   endfor
static void decode_${bitset.get_c_name()}_gen_${bitset.gen_min}(void *out, struct decode_scope *scope)
{
%   if bitset.get_root().decode is not None:
    UNUSED ${bitset.get_root().encode.type} src;
%       if bitset.get_root().encode.type.endswith('*') and name in isa.leafs and bitset.get_root().encode.case_prefix is not None:
    src = ${bitset.get_root().get_c_name()}_create(${s.case_name(bitset.get_root(), bitset.name)});
    *(${bitset.get_root().encode.type} *)out = src;
%       endif
%   endif
}
%endfor

void ${prefix}_isa_disasm(void *bin, int sz, FILE *out, const struct isa_decode_options *options)
{
    isa_disasm(bin, sz, out, options);
}

bool ${prefix}_isa_decode(void *out, void *bin, const struct isa_decode_options *options)
{
    return isa_decode(out, bin, options);
}

uint32_t ${prefix}_isa_get_gpu_id(struct decode_scope *scope)
{
    return isa_get_gpu_id(scope);
}
"""

header = """\
/* Copyright (C) 2020 Google, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the next
 * paragraph) shall be included in all copies or substantial portions of the
 * Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#ifndef _${guard}_
#define _${guard}_

#include "compiler/isaspec/isaspec.h"

#ifdef __cplusplus
extern "C" {
#endif

void ${prefix}_isa_disasm(void *bin, int sz, FILE *out, const struct isa_decode_options *options);
bool ${prefix}_isa_decode(void *out, void *bin, const struct isa_decode_options *options);

struct decode_scope;

uint32_t ${prefix}_isa_get_gpu_id(struct decode_scope *scope);

/**
 * Allows to use gpu_id in expr functions
 */
#define ISA_GPU_ID() ${prefix}_isa_get_gpu_id(scope)

#ifdef __cplusplus
}
#endif

#endif /* _${guard}_ */

"""

def guard(p):
    return os.path.basename(p).upper().replace("-", "_").replace(".", "_")

def prefix(p):
    return os.path.basename(p).lower().replace("-", "_").replace(".", "_").split('_')[0]

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--xml', required=True, help='isaspec XML file.')
    parser.add_argument('--out-c', required=True, help='Output C file.')
    parser.add_argument('--out-h', required=True, help='Output H file.')
    args = parser.parse_args()

    isa = ISA(args.xml)
    s = State(isa)

    try:
        with open(args.out_c, 'w', encoding='utf-8') as f:
            out_h_basename = os.path.basename(args.out_h)
            f.write(Template(template).render(isa=isa, s=s, header=out_h_basename, prefix=prefix(args.out_h)))

        with open(args.out_h, 'w', encoding='utf-8') as f:
            f.write(Template(header).render(isa=isa, guard=guard(args.out_h), prefix=prefix(args.out_h)))

    except Exception:
        # In the event there's an error, this imports some helpers from mako
        # to print a useful stack trace and prints it, then exits with
        # status 1, if python is run with debug; otherwise it just raises
        # the exception
        import sys
        from mako import exceptions
        print(exceptions.text_error_template().render(), file=sys.stderr)
        sys.exit(1)

if __name__ == '__main__':
    main()



import xml.parsers.expat
import sys
import os
import collections

class Error(Exception):
	def __init__(self, message):
		self.message = message

class Enum(object):
	def __init__(self, name):
		self.name = name
		self.values = []

	def has_name(self, name):
		for (n, value) in self.values:
			if n == name:
				return True
		return False

	def dump(self):
		use_hex = False
		for (name, value) in self.values:
			if value > 0x1000:
				use_hex = True

		print("enum %s {" % self.name)
		for (name, value) in self.values:
			if use_hex:
				print("\t%s = 0x%08x," % (name, value))
			else:
				print("\t%s = %d," % (name, value))
		print("};\n")

	def dump_pack_struct(self):
		pass

class Field(object):
	def __init__(self, name, low, high, shr, type, parser):
		self.name = name
		self.low = low
		self.high = high
		self.shr = shr
		self.type = type

		builtin_types = [ None, "a3xx_regid", "boolean", "uint", "hex", "int", "fixed", "ufixed", "float", "address", "waddress" ]

		maxpos = parser.current_bitsize - 1

		if low < 0 or low > maxpos:
			raise parser.error("low attribute out of range: %d" % low)
		if high < 0 or high > maxpos:
			raise parser.error("high attribute out of range: %d" % high)
		if high < low:
			raise parser.error("low is greater than high: low=%d, high=%d" % (low, high))
		if self.type == "boolean" and not low == high:
			raise parser.error("booleans should be 1 bit fields")
		elif self.type == "float" and not (high - low == 31 or high - low == 15):
			raise parser.error("floats should be 16 or 32 bit fields")
		elif not self.type in builtin_types and not self.type in parser.enums:
			raise parser.error("unknown type '%s'" % self.type)

	def ctype(self, var_name):
		if self.type == None:
			type = "uint32_t"
			val = var_name
		elif self.type == "boolean":
			type = "bool"
			val = var_name
		elif self.type == "uint" or self.type == "hex" or self.type == "a3xx_regid":
			type = "uint32_t"
			val = var_name
		elif self.type == "int":
			type = "int32_t"
			val = var_name
		elif self.type == "fixed":
			type = "float"
			val = "((int32_t)(%s * %d.0))" % (var_name, 1 << self.radix)
		elif self.type == "ufixed":
			type = "float"
			val = "((uint32_t)(%s * %d.0))" % (var_name, 1 << self.radix)
		elif self.type == "float" and self.high - self.low == 31:
			type = "float"
			val = "fui(%s)" % var_name
		elif self.type == "float" and self.high - self.low == 15:
			type = "float"
			val = "_mesa_float_to_half(%s)" % var_name
		elif self.type in [ "address", "waddress" ]:
			type = "uint64_t"
			val = var_name
		else:
			type = "enum %s" % self.type
			val = var_name

		if self.shr > 0:
			val = "(%s >> %d)" % (val, self.shr)

		return (type, val)

def tab_to(name, value):
	tab_count = (68 - (len(name) & ~7)) // 8
	if tab_count <= 0:
		tab_count = 1
	print(name + ('\t' * tab_count) + value)

def mask(low, high):
	return ((0xffffffffffffffff >> (64 - (high + 1 - low))) << low)

def field_name(reg, f):
	if f.name:
		name = f.name.lower()
	else:
		# We hit this path when a reg is defined with no bitset fields, ie.
		# 	<reg32 offset="0x88db" name="RB_BLIT_DST_ARRAY_PITCH" low="0" high="28" shr="6" type="uint"/>
		name = reg.name.lower()

	if (name in [ "double", "float", "int" ]) or not (name[0].isalpha()):
			name = "_" + name

	return name

class Bitset(object):
	def __init__(self, name, template):
		self.name = name
		self.inline = False
		if template:
			self.fields = template.fields[:]
		else:
			self.fields = []

	# Get address field if there is one in the bitset, else return None:
	def get_address_field(self):
		for f in self.fields:
			if f.type in [ "address", "waddress" ]:
				return f
		return None

	def dump_regpair_builder(self, reg):
		print("#ifndef NDEBUG")
		known_mask = 0
		for f in self.fields:
			known_mask |= mask(f.low, f.high)
			if f.type in [ "boolean", "address", "waddress" ]:
				continue
			type, val = f.ctype("fields.%s" % field_name(reg, f))
			print("    assert((%-40s & 0x%08x) == 0);" % (val, 0xffffffff ^ mask(0 , f.high - f.low)))
		print("    assert((%-40s & 0x%08x) == 0);" % ("fields.unknown", known_mask))
		print("#endif\n")

		print("    return (struct fd_reg_pair) {")
		if reg.array:
			print("        .reg = REG_%s(__i)," % reg.full_name)
		else:
			print("        .reg = REG_%s," % reg.full_name)

		print("        .value =")
		for f in self.fields:
			if f.type in [ "address", "waddress" ]:
				continue
			else:
				type, val = f.ctype("fields.%s" % field_name(reg, f))
				print("            (%-40s << %2d) |" % (val, f.low))
		value_name = "dword"
		if reg.bit_size == 64:
			value_name = "qword"
		print("            fields.unknown | fields.%s," % (value_name,))

		address = self.get_address_field()
		if address:
			print("        .bo = fields.bo,")
			print("        .is_address = true,")
			if f.type == "waddress":
				print("        .bo_write = true,")
			print("        .bo_offset = fields.bo_offset,")
			print("        .bo_shift = %d," % address.shr)
			print("        .bo_low = %d," % address.low)

		print("    };")

	def dump_pack_struct(self, reg=None):
		if not reg:
			return

		prefix = reg.full_name

		print("struct %s {" % prefix)
		for f in self.fields:
			if f.type in [ "address", "waddress" ]:
				tab_to("    __bo_type", "bo;")
				tab_to("    uint32_t", "bo_offset;")
				continue
			name = field_name(reg, f)

			type, val = f.ctype("var")

			tab_to("    %s" % type, "%s;" % name)
		if reg.bit_size == 64:
			tab_to("    uint64_t", "unknown;")
			tab_to("    uint64_t", "qword;")
		else:
			tab_to("    uint32_t", "unknown;")
			tab_to("    uint32_t", "dword;")
		print("};\n")

		if reg.array:
			print("static inline struct fd_reg_pair\npack_%s(uint32_t __i, struct %s fields)\n{" %
				  (prefix, prefix))
		else:
			print("static inline struct fd_reg_pair\npack_%s(struct %s fields)\n{" %
				  (prefix, prefix))

		self.dump_regpair_builder(reg)

		print("\n}\n")

		if self.get_address_field():
			skip = ", { .reg = 0 }"
		else:
			skip = ""

		if reg.array:
			print("#define %s(__i, ...) pack_%s(__i, __struct_cast(%s) { __VA_ARGS__ })%s\n" %
				  (prefix, prefix, prefix, skip))
		else:
			print("#define %s(...) pack_%s(__struct_cast(%s) { __VA_ARGS__ })%s\n" %
				  (prefix, prefix, prefix, skip))


	def dump(self, prefix=None):
		if prefix == None:
			prefix = self.name
		for f in self.fields:
			if f.name:
				name = prefix + "_" + f.name.upper()
			else:
				name = prefix

			if not f.name and f.low == 0 and f.shr == 0 and not f.type in ["float", "fixed", "ufixed"]:
				pass
			elif f.type == "boolean" or (f.type == None and f.low == f.high):
				tab_to("#define %s" % name, "0x%08x" % (1 << f.low))
			else:
				tab_to("#define %s__MASK" % name, "0x%08x" % mask(f.low, f.high))
				tab_to("#define %s__SHIFT" % name, "%d" % f.low)
				type, val = f.ctype("val")

				print("static inline uint32_t %s(%s val)\n{" % (name, type))
				if f.shr > 0:
					print("\tassert(!(val & 0x%x));" % mask(0, f.shr - 1))
				print("\treturn ((%s) << %s__SHIFT) & %s__MASK;\n}" % (val, name, name))
		print()

class Array(object):
	def __init__(self, attrs, domain, variant):
		if "name" in attrs:
			self.name = attrs["name"]
		else:
			self.name = ""
		self.domain = domain
		self.variant = variant
		self.offset = int(attrs["offset"], 0)
		self.stride = int(attrs["stride"], 0)
		self.length = int(attrs["length"], 0)
		if "usage" in attrs:
			self.usages = attrs["usage"].split(',')
		else:
			self.usages = None

	def dump(self):
		print("#define _%s(i0) (0x%08x + 0x%x*(i0))\n" % (self.name, self.offset, self.stride))

	def dump_pack_struct(self):
		pass

	def dump_regpair_builder(self):
		pass

class Reg(object):
	def __init__(self, attrs, domain, array, bit_size):
		self.name = attrs["name"]
		self.domain = domain
		self.array = array
		self.offset = int(attrs["offset"], 0)
		self.type = None
		self.bit_size = bit_size
		if array:
			self.name = array.name + "_" + self.name
		self.full_name = self.name

	def dump(self):
		if self.array:
			offset = self.array.offset + self.offset
			print("static inline uint32_t %s(uint32_t i0) { return 0x%08x + 0x%x*i0; }" % (self.full_name, offset, self.array.stride))
		else:
			tab_to("#define %s" % self.full_name, "0x%08x" % self.offset)

		if self.bitset.inline:
			self.bitset.dump(self.full_name)

	def dump_pack_struct(self):
		if self.bitset.inline:
			self.bitset.dump_pack_struct(self)

	def dump_regpair_builder(self):
		if self.bitset.inline:
			self.bitset.dump_regpair_builder(self)

	def dump_py(self):
		print("\tREG_%s = 0x%08x" % (self.full_name, self.offset))


class Parser(object):
	def __init__(self):
		self.current_array = None
		self.current_domain = None
		self.current_prefix = None
		self.current_prefix_type = None
		self.current_stripe = None
		self.current_bitset = None
		self.current_bitsize = 32
		# The varset attribute on the domain specifies the enum which
		# specifies all possible hw variants:
		self.current_varset = None
		# Regs that have multiple variants.. we only generated the C++
		# template based struct-packers for these
		self.variant_regs = {}
		# Information in which contexts regs are used, to be used in
		# debug options
		self.usage_regs = collections.defaultdict(list)
		self.bitsets = {}
		self.enums = {}
		self.variants = set()
		self.file = []
		self.xml_files = []
		self.copyright_year = None
		self.authors = []
		self.license = None

	def error(self, message):
		parser, filename = self.stack[-1]
		return Error("%s:%d:%d: %s" % (filename, parser.CurrentLineNumber, parser.CurrentColumnNumber, message))

	def prefix(self, variant=None):
		if self.current_prefix_type == "variant" and variant:
			return variant
		elif self.current_stripe:
			return self.current_stripe + "_" + self.current_domain
		elif self.current_prefix:
			return self.current_prefix + "_" + self.current_domain
		else:
			return self.current_domain

	def parse_field(self, name, attrs):
		try:
			if "pos" in attrs:
				high = low = int(attrs["pos"], 0)
			elif "high" in attrs and "low" in attrs:
				high = int(attrs["high"], 0)
				low = int(attrs["low"], 0)
			else:
				low = 0
				high = self.current_bitsize - 1

			if "type" in attrs:
				type = attrs["type"]
			else:
				type = None

			if "shr" in attrs:
				shr = int(attrs["shr"], 0)
			else:
				shr = 0

			b = Field(name, low, high, shr, type, self)

			if type == "fixed" or type == "ufixed":
				b.radix = int(attrs["radix"], 0)

			self.current_bitset.fields.append(b)
		except ValueError as e:
			raise self.error(e)

	def parse_varset(self, attrs):
		# Inherit the varset from the enclosing domain if not overriden:
		varset = self.current_varset
		if "varset" in attrs:
			varset = self.enums[attrs["varset"]]
		return varset

	def parse_variants(self, attrs):
		if not "variants" in attrs:
				return None
		variant = attrs["variants"].split(",")[0]
		if "-" in variant:
			variant = variant[:variant.index("-")]

		varset = self.parse_varset(attrs)

		assert varset.has_name(variant)

		return variant

	def add_all_variants(self, reg, attrs, parent_variant):
		# TODO this should really handle *all* variants, including dealing
		# with open ended ranges (ie. "A2XX,A4XX-") (we have the varset
		# enum now to make that possible)
		variant = self.parse_variants(attrs)
		if not variant:
			variant = parent_variant

		if reg.name not in self.variant_regs:
			self.variant_regs[reg.name] = {}
		else:
			# All variants must be same size:
			v = next(iter(self.variant_regs[reg.name]))
			assert self.variant_regs[reg.name][v].bit_size == reg.bit_size

		self.variant_regs[reg.name][variant] = reg

	def add_all_usages(self, reg, usages):
		if not usages:
			return

		for usage in usages:
			self.usage_regs[usage].append(reg)

		self.variants.add(reg.domain)

	def do_validate(self, schemafile):
		try:
			from lxml import etree

			parser, filename = self.stack[-1]
			dirname = os.path.dirname(filename)

			# we expect this to look like <namespace url> schema.xsd.. I think
			# technically it is supposed to be just a URL, but that doesn't
			# quite match up to what we do.. Just skip over everything up to
			# and including the first whitespace character:
			schemafile = schemafile[schemafile.rindex(" ")+1:]

			# this is a bit cheezy, but the xml file to validate could be
			# in a child director, ie. we don't really know where the schema
			# file is, the way the rnn C code does.  So if it doesn't exist
			# just look one level up
			if not os.path.exists(dirname + "/" + schemafile):
				schemafile = "../" + schemafile

			if not os.path.exists(dirname + "/" + schemafile):
				raise self.error("Cannot find schema for: " + filename)

			xmlschema_doc = etree.parse(dirname + "/" + schemafile)
			xmlschema = etree.XMLSchema(xmlschema_doc)

			xml_doc = etree.parse(filename)
			if not xmlschema.validate(xml_doc):
				error_str = str(xmlschema.error_log.filter_from_errors()[0])
				raise self.error("Schema validation failed for: " + filename + "\n" + error_str)
		except ImportError:
			print("lxml not found, skipping validation", file=sys.stderr)

	def do_parse(self, filename):
		filepath = os.path.abspath(filename)
		if filepath in self.xml_files:
			return
		self.xml_files.append(filepath)
		file = open(filename, "rb")
		parser = xml.parsers.expat.ParserCreate()
		self.stack.append((parser, filename))
		parser.StartElementHandler = self.start_element
		parser.EndElementHandler = self.end_element
		parser.CharacterDataHandler = self.character_data
		parser.buffer_text = True
		parser.ParseFile(file)
		self.stack.pop()
		file.close()

	def parse(self, rnn_path, filename):
		self.path = rnn_path
		self.stack = []
		self.do_parse(filename)

	def parse_reg(self, attrs, bit_size):
		self.current_bitsize = bit_size
		if "type" in attrs and attrs["type"] in self.bitsets:
			bitset = self.bitsets[attrs["type"]]
			if bitset.inline:
				self.current_bitset = Bitset(attrs["name"], bitset)
				self.current_bitset.inline = True
			else:
				self.current_bitset = bitset
		else:
			self.current_bitset = Bitset(attrs["name"], None)
			self.current_bitset.inline = True
			if "type" in attrs:
				self.parse_field(None, attrs)

		variant = self.parse_variants(attrs)
		if not variant and self.current_array:
			variant = self.current_array.variant

		self.current_reg = Reg(attrs, self.prefix(variant), self.current_array, bit_size)
		self.current_reg.bitset = self.current_bitset

		if len(self.stack) == 1:
			self.file.append(self.current_reg)

		if variant is not None:
			self.add_all_variants(self.current_reg, attrs, variant)

		usages = None
		if "usage" in attrs:
			usages = attrs["usage"].split(',')
		elif self.current_array:
			usages = self.current_array.usages

		self.add_all_usages(self.current_reg, usages)

	def start_element(self, name, attrs):
		self.cdata = ""
		if name == "import":
			filename = attrs["file"]
			self.do_parse(os.path.join(self.path, filename))
		elif name == "domain":
			self.current_domain = attrs["name"]
			if "prefix" in attrs:
				self.current_prefix = self.parse_variants(attrs)
				self.current_prefix_type = attrs["prefix"]
			else:
				self.current_prefix = None
				self.current_prefix_type = None
			if "varset" in attrs:
				self.current_varset = self.enums[attrs["varset"]]
		elif name == "stripe":
			self.current_stripe = self.parse_variants(attrs)
		elif name == "enum":
			self.current_enum_value = 0
			self.current_enum = Enum(attrs["name"])
			self.enums[attrs["name"]] = self.current_enum
			if len(self.stack) == 1:
				self.file.append(self.current_enum)
		elif name == "value":
			if "value" in attrs:
				value = int(attrs["value"], 0)
			else:
				value = self.current_enum_value
			self.current_enum.values.append((attrs["name"], value))
		elif name == "reg32":
			self.parse_reg(attrs, 32)
		elif name == "reg64":
			self.parse_reg(attrs, 64)
		elif name == "array":
			self.current_bitsize = 32
			variant = self.parse_variants(attrs)
			self.current_array = Array(attrs, self.prefix(variant), variant)
			if len(self.stack) == 1:
				self.file.append(self.current_array)
		elif name == "bitset":
			self.current_bitset = Bitset(attrs["name"], None)
			if "inline" in attrs and attrs["inline"] == "yes":
				self.current_bitset.inline = True
			self.bitsets[self.current_bitset.name] = self.current_bitset
			if len(self.stack) == 1 and not self.current_bitset.inline:
				self.file.append(self.current_bitset)
		elif name == "bitfield" and self.current_bitset:
			self.parse_field(attrs["name"], attrs)
		elif name == "database":
			self.do_validate(attrs["xsi:schemaLocation"])
		elif name == "copyright":
			self.copyright_year = attrs["year"]
		elif name == "author":
			self.authors.append(attrs["name"] + " <" + attrs["email"] + "> " + attrs["name"])

	def end_element(self, name):
		if name == "domain":
			self.current_domain = None
			self.current_prefix = None
			self.current_prefix_type = None
		elif name == "stripe":
			self.current_stripe = None
		elif name == "bitset":
			self.current_bitset = None
		elif name == "reg32":
			self.current_reg = None
		elif name == "array":
			self.current_array = None
		elif name == "enum":
			self.current_enum = None
		elif name == "license":
			self.license = self.cdata

	def character_data(self, data):
		self.cdata += data

	def dump_reg_usages(self):
		d = collections.defaultdict(list)
		for usage, regs in self.usage_regs.items():
			for reg in regs:
				variants = self.variant_regs.get(reg.name)
				if variants:
					for variant, vreg in variants.items():
						if reg == vreg:
							d[(usage, variant)].append(reg)
				else:
					for variant in self.variants:
						d[(usage, variant)].append(reg)

		print("#ifdef __cplusplus")

		for usage, regs in self.usage_regs.items():
			print("template<chip CHIP> constexpr inline uint16_t %s_REGS[] = {};" % (usage.upper()))

		for (usage, variant), regs in d.items():
			offsets = []

			for reg in regs:
				if reg.array:
					for i in range(reg.array.length):
						offsets.append(reg.array.offset + reg.offset + i * reg.array.stride)
						if reg.bit_size == 64:
							offsets.append(offsets[-1] + 1)
				else:
					offsets.append(reg.offset)
					if reg.bit_size == 64:
						offsets.append(offsets[-1] + 1)

			offsets.sort()

			print("template<> constexpr inline uint16_t %s_REGS<%s>[] = {" % (usage.upper(), variant))
			for offset in offsets:
				print("\t%s," % hex(offset))
			print("};")

		print("#endif")

	def dump(self):
		enums = []
		bitsets = []
		regs = []
		for e in self.file:
			if isinstance(e, Enum):
				enums.append(e)
			elif isinstance(e, Bitset):
				bitsets.append(e)
			else:
				regs.append(e)

		for e in enums + bitsets + regs:
			e.dump()

		self.dump_reg_usages()

		print("static inline char *ethosu_get_cmd_name(unsigned domain, uint32_t cmd) {")
		for e in regs:
			if e.array:
				continue
			domain = 0 if e.domain == "CMD0" else 1
			print("    if (domain == %d && cmd == 0x%08x) return \"%s\";" % (domain, e.offset, e.full_name))
		print("    return NULL;")
		print("}\n")

	def dump_regs_py(self):
		regs = []
		for e in self.file:
			if isinstance(e, Reg):
				regs.append(e)

		for e in regs:
			e.dump_py()


	def dump_reg_variants(self, regname, variants):
		# Don't bother for things that only have a single variant:
		if len(variants) == 1:
			return
		print("#ifdef __cplusplus")
		print("struct __%s {" % regname)
		# TODO be more clever.. we should probably figure out which
		# fields have the same type in all variants (in which they
		# appear) and stuff everything else in a variant specific
		# sub-structure.
		seen_fields = []
		bit_size = 32
		array = False
		address = None
		for variant in variants.keys():
			print("    /* %s fields: */" % variant)
			reg = variants[variant]
			bit_size = reg.bit_size
			array = reg.array
			for f in reg.bitset.fields:
				fld_name = field_name(reg, f)
				if fld_name in seen_fields:
					continue
				seen_fields.append(fld_name)
				name = fld_name.lower()
				if f.type in [ "address", "waddress" ]:
					if address:
						continue
					address = f
					tab_to("    __bo_type", "bo;")
					tab_to("    uint32_t", "bo_offset;")
					continue
				type, val = f.ctype("var")
				tab_to("    %s" %type, "%s;" %name)
		print("    /* fallback fields: */")
		if bit_size == 64:
			tab_to("    uint64_t", "unknown;")
			tab_to("    uint64_t", "qword;")
		else:
			tab_to("    uint32_t", "unknown;")
			tab_to("    uint32_t", "dword;")
		print("};")
		# TODO don't hardcode the varset enum name
		varenum = "chip"
		print("template <%s %s>" % (varenum, varenum.upper()))
		print("static inline struct fd_reg_pair")
		xtra = ""
		xtravar = ""
		if array:
			xtra = "int __i, "
			xtravar = "__i, "
		print("__%s(%sstruct __%s fields) {" % (regname, xtra, regname))
		for variant in variants.keys():
			print("  if (%s == %s) {" % (varenum.upper(), variant))
			reg = variants[variant]
			reg.dump_regpair_builder()
			print("  } else")
		print("    assert(!\"invalid variant\");")
		print("}")

		if bit_size == 64:
			skip = ", { .reg = 0 }"
		else:
			skip = ""

		print("#define %s(VARIANT, %s...) __%s<VARIANT>(%s{__VA_ARGS__})%s" % (regname, xtravar, regname, xtravar, skip))
		print("#endif /* __cplusplus */")

	def dump_structs(self):
		for e in self.file:
			e.dump_pack_struct()

		for regname in self.variant_regs:
			self.dump_reg_variants(regname, self.variant_regs[regname])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# The Mesa 3D Graphics Library documentation build configuration file, created by
# sphinx-quickstart on Wed Mar 29 14:08:51 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

from hawkmoth.util import compiler

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.append(os.path.abspath('_exts'))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'bootstrap',
    'depfile',
    'formatting',
    'hawkmoth',
    'nir',
    'redirects',
    'sphinx.ext.graphviz',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'The Mesa 3D Graphics Library'
copyright = '1995-2018, Brian Paul'
author = 'Brian Paul'
html_show_copyright = False

html_theme_path = ['.']

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = 'latest'
# The full version, including alpha/beta/rc tags.
release = 'latest'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['header-stubs']

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# Disable highlighting unless a language is specified, otherwise we'll get
# python keywords highlit in literal blocks.
highlight_language = 'none'

default_role = 'c:expr'

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'mesa3d_theme'

html_favicon = 'favicon.ico'

html_copy_source = False

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

html_extra_path = [
    '_extra/',
    'release-maintainers-keys.asc',
    'features.txt',
    'libGL.txt',
    'README.UVD',
    'README.VCE',
]

html_redirects = [
    ('webmaster', 'https://www.mesa3d.org/website/'),
    ('developers', 'https://www.mesa3d.org/developers/'),
    ('thanks', 'https://gitlab.freedesktop.org/mesa/mesa/-/blob/amber/docs/thanks.rst'),
]


# -- Options for linkcheck ------------------------------------------------

linkcheck_ignore = [
    r'specs/.*\.spec',  # gets copied during the build process
    r'news:.*',  # seems linkcheck doesn't like the news: URI-scheme...
    r'http://mesa-ci-results.jf.intel.com',  # only available for Intel employees
    r'https://gitlab.com/.*#.*',  # needs JS eval
    r'https://gitlab.freedesktop.org/.*#.*',  # needs JS eval
    r'https://github.com/.*#.*',  # needs JS eval
    r'https://www.intel.com/.*',  # intel.com is blocking the linkcheck user-agent; maybe it can be customized to look like a browser?
    r'https://sourceforge.net/.*',  # blocking the linkcheck user-agent
    r'https://.*\.sourceforge\.net/.*',  # blocking the linkcheck user-agent
    r'https://stackoverflow.com/.*',  # blocking the linkcheck user-agent
    r'https://(www|dev)\.vulkan\.org/.*',  # blocking the linkcheck user-agent
    r'https://crates.io/.*',  # blocking the linkcheck user-agent
]
linkcheck_exclude_documents = [r'relnotes/.*']

linkcheck_allowed_redirects = {
    # Pages that forward the front-page to a wiki or some explore-page
    'https://www.freedesktop.org': 'https://www.freedesktop.org/wiki/',
    'https://x.org': 'https://x.org/wiki/',
    'https://dri.freedesktop.org/': 'https://dri.freedesktop.org/wiki/',
    'https://gitlab.freedesktop.org/': 'https://gitlab.freedesktop.org/explore/groups',
    'https://www.sphinx-doc.org/': 'https://www.sphinx-doc.org/en/master/',

    # Pages that requires authentication
    'https://gitlab.freedesktop.org/admin/runners': 'https://gitlab.freedesktop.org/users/sign_in',
    'https://gitlab.freedesktop.org/profile/personal_access_tokens': 'https://gitlab.freedesktop.org/users/sign_in',
    'https://support.broadcom.com/group/ecx/free-downloads': 'https://support.broadcom.com/c/portal/login',
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'TheMesa3DGraphicsLibrarydoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'TheMesa3DGraphicsLibrary.tex', 'The Mesa 3D Graphics Library Documentation',
     'Brian Paul', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'themesa3dgraphicslibrary', 'The Mesa 3D Graphics Library Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'TheMesa3DGraphicsLibrary', 'The Mesa 3D Graphics Library Documentation',
     author, 'TheMesa3DGraphicsLibrary', 'One line description of project.',
     'Miscellaneous'),
]

# -- Options for Graphviz -------------------------------------------------

graphviz_output_format = 'svg'

# -- Options for hawkmoth -------------------------------------------------

hawkmoth_root = os.path.abspath(os.pardir)
mesa_root = os.path.join(os.path.dirname(__file__), os.pardir)
mesa_build_root = os.environ.get('MESA_BUILD_ROOT')
hawkmoth_clang = [
    '-I{}/docs/header-stubs/'.format(mesa_root),
    '-I{}/include/'.format(mesa_root),
    '-I{}/src/'.format(mesa_root),
    '-I{}/src/gallium/include/'.format(mesa_root),
    '-I{}/src/intel/'.format(mesa_root),
    '-I{}/src/mesa/'.format(mesa_root),
    '-I{}/src/vulkan/util'.format(mesa_root),
    '-I{}/src/'.format(mesa_build_root),
    '-DHAVE_STRUCT_TIMESPEC',
    '-DHAVE_PTHREAD',
    '-DHAVE_ENDIAN_H',
]
hawkmoth_clang.extend(compiler.get_include_args())

# helpers for definining parameter direction
rst_prolog = '''
.. |in| replace:: **[in]**
.. |out| replace:: **[out]**
.. |inout| replace:: **[inout]**
'''



# Copyright © 2013 Intel Corporation
# SPDX-License-Identifier: MIT

import sys

from builtin_types import BUILTIN_TYPES
from mako.template import Template

template = """\
/*
 * Copyright 2023 Intel Corporation
 * SPDX-License-Identifier: MIT
 */

/* This is an automatically generated file. */

#include "glsl_types.h"
#include "util/glheader.h"

const char glsl_type_builtin_names[] =
%for n in NAME_ARRAY:
   "${n}"
%endfor
;

%for t in BUILTIN_TYPES:
const struct glsl_type glsl_type_builtin_${t["name"]} = {
   %for k, v in t.items():
       %if v is None or k == "name":
          <% continue %>
       %elif k == "name_id":
          .name_id = ${v},
          .has_builtin_name = 1,
       %else:
          .${k} = ${v},
       %endif
   %endfor
};

%endfor"""

if len(sys.argv) < 2:
    print('Missing output argument', file=sys.stderr)
    sys.exit(1)

output = sys.argv[1]

# Add padding to make sure zero is an invalid name.
invalid = "INVALID"
NAME_ARRAY = [invalid + "\\0"]
id = len(invalid) + 1

for t in BUILTIN_TYPES:
    name = t["name"]
    NAME_ARRAY.append(name + "\\0")
    t["name_id"] = id
    id += len(name) + 1

with open(output, 'w', encoding='utf-8') as f:
    f.write(Template(template).render(BUILTIN_TYPES=BUILTIN_TYPES, NAME_ARRAY=NAME_ARRAY))



# Copyright © 2017 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Create enum to string functions for vulkan using vk.xml."""

import argparse
import functools
import os
import re
import textwrap
import xml.etree.ElementTree as et

from mako.template import Template
from vk_extensions import Extension, filter_api, get_all_required

COPYRIGHT = textwrap.dedent(u"""\
    * Copyright © 2017 Intel Corporation
    *
    * Permission is hereby granted, free of charge, to any person obtaining a copy
    * of this software and associated documentation files (the "Software"), to deal
    * in the Software without restriction, including without limitation the rights
    * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    * copies of the Software, and to permit persons to whom the Software is
    * furnished to do so, subject to the following conditions:
    *
    * The above copyright notice and this permission notice shall be included in
    * all copies or substantial portions of the Software.
    *
    * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    * SOFTWARE.""")

C_TEMPLATE = Template(textwrap.dedent(u"""\
    /* Autogenerated file -- do not edit
     * generated by ${file}
     *
     ${copyright}
     */

    #include <string.h>
    #include <vulkan/vulkan_core.h>
    #include <vulkan/vk_android_native_buffer.h>
    #include <vulkan/vk_layer.h>
    #include "util/macros.h"
    #include "vk_enum_to_str.h"

    % for enum in enums:

      % if enum.guard:
#ifdef ${enum.guard}
      % endif
    const char *
    vk_${enum.name[2:]}_to_str(${enum.name} input)
    {
        switch((int64_t)input) {
    % for v in sorted(enum.values.keys()):
        case ${v}:
            return "${enum.values[v]}";
    % endfor
        case ${enum.max_enum_name}: return "${enum.max_enum_name}";
        default:
            return "Unknown ${enum.name} value.";
        }
    }

      % if enum.guard:
#endif
      % endif
    %endfor

    % for enum in bitmasks:

      % if enum.guard:
#ifdef ${enum.guard}
      % endif
    const char *
    vk_${enum.name[2:]}_to_str(${enum.name} input)
    {
        switch((int64_t)input) {
    % for v in sorted(enum.values.keys()):
        case ${v}:
            return "${enum.values[v]}";
    % endfor
        default:
            return "Unknown ${enum.name} value.";
        }
    }

      % if enum.guard:
#endif
      % endif
    %endfor

    size_t vk_structure_type_size(const struct VkBaseInStructure *item)
    {
        switch((int)item->sType) {
    % for struct in structs:
        % if struct.extension is not None and struct.extension.define is not None:
    #ifdef ${struct.extension.define}
        case ${struct.stype}: return sizeof(${struct.name});
    #endif
        % else:
        case ${struct.stype}: return sizeof(${struct.name});
        % endif
    %endfor
        case VK_STRUCTURE_TYPE_LOADER_INSTANCE_CREATE_INFO: return sizeof(VkLayerInstanceCreateInfo);
        case VK_STRUCTURE_TYPE_LOADER_DEVICE_CREATE_INFO: return sizeof(VkLayerDeviceCreateInfo);
        default:
            UNREACHABLE("Undefined struct type.");
        }
    }

    const char *
    vk_ObjectType_to_ObjectName(VkObjectType type)
    {
        switch((int)type) {
    % for object_type in sorted(object_types[0].enum_to_name.keys()):
        case ${object_type}:
            return "${object_types[0].enum_to_name[object_type]}";
    % endfor
        default:
            return "Unknown VkObjectType value.";
        }
    }
    """))

H_TEMPLATE = Template(textwrap.dedent(u"""\
    /* Autogenerated file -- do not edit
     * generated by ${file}
     *
     ${copyright}
     */

    #ifndef MESA_VK_ENUM_TO_STR_H
    #define MESA_VK_ENUM_TO_STR_H

    #include <vulkan/vulkan.h>
    #include <vulkan/vk_android_native_buffer.h>

    #ifdef __cplusplus
    extern "C" {
    #endif

    % for enum in enums:
      % if enum.guard:
#ifdef ${enum.guard}
      % endif
    const char * vk_${enum.name[2:]}_to_str(${enum.name} input);
      % if enum.guard:
#endif
      % endif
    % endfor

    % for enum in bitmasks:
      % if enum.guard:
#ifdef ${enum.guard}
      % endif
    const char * vk_${enum.name[2:]}_to_str(${enum.name} input);
      % if enum.guard:
#endif
      % endif
    % endfor

    size_t vk_structure_type_size(const struct VkBaseInStructure *item);

    const char * vk_ObjectType_to_ObjectName(VkObjectType type);

    #ifdef __cplusplus
    } /* extern "C" */
    #endif

    #endif
    """))


H_DEFINE_TEMPLATE = Template(textwrap.dedent(u"""\
    /* Autogenerated file -- do not edit
     * generated by ${file}
     *
     ${copyright}
     */

    #ifndef MESA_VK_ENUM_DEFINES_H
    #define MESA_VK_ENUM_DEFINES_H

    #include <vulkan/vulkan_core.h>
    #include <vulkan/vk_android_native_buffer.h>

    #ifdef __cplusplus
    extern "C" {
    #endif

    % for ext in extensions:
    #define _${ext.name}_number (${ext.number})
    % endfor

    % for enum in bitmasks:
      % if enum.bitwidth > 32:
        <% continue %>
      % endif
      % if enum.guard:
#ifdef ${enum.guard}
      % endif
    #define ${enum.all_bits_name()} ${hex(enum.all_bits_value())}u
      % if enum.guard:
#endif
      % endif
    % endfor

    % for enum in bitmasks:
      % if enum.bitwidth < 64:
        <% continue %>
      % endif
    /* Redefine bitmask values of ${enum.name} */
      % if enum.guard:
#ifdef ${enum.guard}
      % endif
      % for n, v in enum.name_to_value.items():
    #define ${n} (${hex(v)}ULL)
      % endfor
      % if enum.guard:
#endif
      % endif
    % endfor

    static inline VkFormatFeatureFlags
    vk_format_features2_to_features(VkFormatFeatureFlags2 features2)
    {
       return features2 & VK_ALL_FORMAT_FEATURE_FLAG_BITS;
    }

    #ifdef __cplusplus
    } /* extern "C" */
    #endif

    #endif
    """))


class NamedFactory(object):
    """Factory for creating enums."""

    def __init__(self, type_):
        self.registry = {}
        self.type = type_

    def __call__(self, name, **kwargs):
        try:
            return self.registry[name]
        except KeyError:
            n = self.registry[name] = self.type(name, **kwargs)
        return n

    def get(self, name):
        return self.registry.get(name)


class VkExtension(object):
    """Simple struct-like class representing extensions"""

    def __init__(self, name, number=None, define=None):
        self.name = name
        self.number = number
        self.define = define


def CamelCase_to_SHOUT_CASE(s):
   return (s[:1] + re.sub(r'(?<![A-Z])([A-Z])', r'_\1', s[1:])).upper()

def compute_max_enum_name(s):
    if s == "VkSwapchainImageUsageFlagBitsANDROID":
        return "VK_SWAPCHAIN_IMAGE_USAGE_FLAG_BITS_MAX_ENUM"
    if s == "VkTensorTilingARM":
        return "VK_TENSOR_TILING_MAX_ENUM_ARM"
    max_enum_name = CamelCase_to_SHOUT_CASE(s)
    last_prefix = max_enum_name.rsplit('_', 1)[-1]
    # Those special prefixes need to be always at the end
    if last_prefix in ['AMD', 'AMDX', 'EXT', 'INTEL', 'KHR', 'NV', 'LUNARG', 'QCOM', 'MSFT', 'ARM'] :
        max_enum_name = "_".join(max_enum_name.split('_')[:-1])
        max_enum_name = max_enum_name + "_MAX_ENUM_" + last_prefix
    else:
        max_enum_name = max_enum_name + "_MAX_ENUM"

    return max_enum_name

class VkEnum(object):
    """Simple struct-like class representing a single Vulkan Enum."""

    def __init__(self, name, bitwidth=32, values=None):
        self.name = name
        self.max_enum_name = compute_max_enum_name(name)
        self.bitwidth = bitwidth
        self.extension = None
        # Maps numbers to names
        self.values = values or dict()
        self.name_to_value = dict()
        self.guard = None
        self.name_to_alias_list = {}

    def all_bits_name(self):
        assert self.name.startswith('Vk')
        assert re.search(r'FlagBits[A-Z]*$', self.name)

        return 'VK_ALL_' + CamelCase_to_SHOUT_CASE(self.name[2:])

    def all_bits_value(self):
        return functools.reduce(lambda a,b: a | b, self.values.keys(), 0)

    def add_value(self, name, value=None,
                  extnum=None, offset=None, alias=None,
                  error=False):
        if alias is not None:
            assert value is None and offset is None
            if alias not in self.name_to_value:
                # We don't have this alias yet.  Just record the alias and
                # we'll deal with it later.
                alias_list = self.name_to_alias_list.setdefault(alias, [])
                alias_list.append(name);
                return

            # Use the value from the alias
            value = self.name_to_value[alias]

        assert value is not None or extnum is not None
        if value is None:
            value = 1000000000 + (extnum - 1) * 1000 + offset
            if error:
                value = -value

        self.name_to_value[name] = value
        if value not in self.values:
            self.values[value] = name
        elif len(self.values[value]) > len(name):
            self.values[value] = name

        # Now that the value has been fully added, resolve aliases, if any.
        if name in self.name_to_alias_list:
            for alias in self.name_to_alias_list[name]:
                self.add_value(alias, value)
            del self.name_to_alias_list[name]

    def add_value_from_xml(self, elem, extension=None):
        self.extension = extension
        if 'value' in elem.attrib:
            self.add_value(elem.attrib['name'],
                           value=int(elem.attrib['value'], base=0))
        elif 'bitpos' in elem.attrib:
            self.add_value(elem.attrib['name'],
                           value=(1 << int(elem.attrib['bitpos'], base=0)))
        elif 'alias' in elem.attrib:
            self.add_value(elem.attrib['name'], alias=elem.attrib['alias'])
        else:
            error = 'dir' in elem.attrib and elem.attrib['dir'] == '-'
            if 'extnumber' in elem.attrib:
                extnum = int(elem.attrib['extnumber'])
            else:
                extnum = extension.number
            self.add_value(elem.attrib['name'],
                           extnum=extnum,
                           offset=int(elem.attrib['offset']),
                           error=error)

    def set_guard(self, g):
        self.guard = g


class VkChainStruct(object):
    """Simple struct-like class representing a single Vulkan struct identified with a VkStructureType"""
    def __init__(self, name, stype):
        self.name = name
        self.stype = stype
        self.extension = None


def struct_get_stype(xml_node):
    for member in xml_node.findall('./member'):
        name = member.findall('./name')
        if len(name) > 0 and name[0].text == "sType":
            return member.get('values')
    return None

class VkObjectType(object):
    """Simple struct-like class representing a single Vulkan object type"""
    def __init__(self, name):
        self.name = name
        self.enum_to_name = dict()


def parse_xml(enum_factory, ext_factory, struct_factory, bitmask_factory,
              obj_type_factory, filename, beta):
    """Parse the XML file. Accumulate results into the factories.

    This parser is a memory efficient iterative XML parser that returns a list
    of VkEnum objects.
    """

    xml = et.parse(filename)
    api = 'vulkan'

    required_types = get_all_required(xml, 'type', api, beta)

    for enum_type in xml.findall('./enums[@type="enum"]'):
        if not filter_api(enum_type, api):
            continue

        type_name = enum_type.attrib['name']
        if not type_name in required_types:
            continue

        enum = enum_factory(type_name)
        for value in enum_type.findall('./enum'):
            if filter_api(value, api):
                enum.add_value_from_xml(value)

    # For bitmask we only add the Enum selected for convenience.
    for enum_type in xml.findall('./enums[@type="bitmask"]'):
        if not filter_api(enum_type, api):
            continue

        type_name = enum_type.attrib['name']
        if not type_name in required_types:
            continue

        bitwidth = int(enum_type.attrib.get('bitwidth', 32))
        enum = bitmask_factory(type_name, bitwidth=bitwidth)
        for value in enum_type.findall('./enum'):
            if filter_api(value, api):
                enum.add_value_from_xml(value)

    for feature in xml.findall('./feature'):
        if not api in feature.attrib['api'].split(','):
            continue

        for value in feature.findall('./require/enum[@extends]'):
            extends = value.attrib['extends']
            enum = enum_factory.get(extends)
            if enum is not None:
                enum.add_value_from_xml(value)
            enum = bitmask_factory.get(extends)
            if enum is not None:
                enum.add_value_from_xml(value)

    for struct_type in xml.findall('./types/type[@category="struct"]'):
        if not filter_api(struct_type, api):
            continue

        name = struct_type.attrib['name']
        if name not in required_types:
            continue

        stype = struct_get_stype(struct_type)
        if stype is not None:
            struct_factory(name, stype=stype)

    platform_define = {}
    for platform in xml.findall('./platforms/platform'):
        name = platform.attrib['name']
        define = platform.attrib['protect']
        platform_define[name] = define

    for ext_elem in xml.findall('./extensions/extension'):
        ext = Extension.from_xml(ext_elem)
        if api not in ext.supported:
            continue

        define = platform_define.get(ext.platform, None)
        extension = ext_factory(ext.name, number=ext.number, define=define)

        for req_elem in ext_elem.findall('./require'):
            if not filter_api(req_elem, api):
                continue

            for value in req_elem.findall('./enum[@extends]'):
                extends = value.attrib['extends']
                enum = enum_factory.get(extends)
                if enum is not None:
                    enum.add_value_from_xml(value, extension)
                enum = bitmask_factory.get(extends)
                if enum is not None:
                    enum.add_value_from_xml(value, extension)

            for t in req_elem.findall('./type'):
                struct = struct_factory.get(t.attrib['name'])
                if struct is not None:
                    struct.extension = extension

        if define:
            for value in ext_elem.findall('./require/type[@name]'):
                enum = enum_factory.get(value.attrib['name'])
                if enum is not None:
                    enum.set_guard(define)
                enum = bitmask_factory.get(value.attrib['name'])
                if enum is not None:
                    enum.set_guard(define)

    obj_type_enum = enum_factory.get("VkObjectType")
    obj_types = obj_type_factory("VkObjectType")
    for object_type in xml.findall('./types/type[@category="handle"]'):
        for object_name in object_type.findall('./name'):
            # Convert to int to avoid undefined enums
            enum = object_type.attrib['objtypeenum']

            # Annoyingly, object types are hard to filter by API so just
            # look for whether or not we can find the enum name in the
            # VkObjectType enum.
            if enum not in obj_type_enum.name_to_value:
                continue

            enum_val = obj_type_enum.name_to_value[enum]
            obj_types.enum_to_name[enum_val] = object_name.text


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--beta', required=True, help='Enable beta extensions.')
    parser.add_argument('--xml', required=True,
                        help='Vulkan API XML files',
                        action='append',
                        dest='xml_files')
    parser.add_argument('--out-c',
                        help='Output C file',
                        required=True)
    parser.add_argument('--out-h',
                        help='Output H file',
                        required=True)
    parser.add_argument('--out-d',
                        help='Output defines H file',
                        required=True)

    args = parser.parse_args()

    enum_factory = NamedFactory(VkEnum)
    ext_factory = NamedFactory(VkExtension)
    struct_factory = NamedFactory(VkChainStruct)
    obj_type_factory = NamedFactory(VkObjectType)
    bitmask_factory = NamedFactory(VkEnum)

    for filename in args.xml_files:
        parse_xml(enum_factory, ext_factory, struct_factory, bitmask_factory,
                  obj_type_factory, filename, args.beta)
    enums = sorted(enum_factory.registry.values(), key=lambda e: e.name)
    extensions = sorted(ext_factory.registry.values(), key=lambda e: e.name)
    structs = sorted(struct_factory.registry.values(), key=lambda e: e.name)
    bitmasks = sorted(bitmask_factory.registry.values(), key=lambda e: e.name)
    object_types = sorted(obj_type_factory.registry.values(), key=lambda e: e.name)

    for template, file_ in [(C_TEMPLATE, args.out_c),
                            (H_TEMPLATE, args.out_h),
                            (H_DEFINE_TEMPLATE, args.out_d)]:
        with open(file_, 'w', encoding='utf-8') as f:
            f.write(template.render(
                file=os.path.basename(__file__),
                enums=enums,
                extensions=extensions,
                structs=structs,
                bitmasks=bitmasks,
                object_types=object_types,
                copyright=COPYRIGHT))


if __name__ == '__main__':
    main()



#! /usr/bin/env python3

import argparse
from collections import defaultdict
from mako.template import Template
import util

TEMPLATE_H = Template(
    """\
#pragma once

%for (engine_id, classes) in classes_per_eng.items():
%for cl in classes:
#include "nv_push_cl${cl_for_filename(cl)}.h"
%endfor
%endfor


static inline const char*
P_PARSE_NV_MTHD(uint16_t class_id, uint16_t idx)
{
    uint16_t cls_hi = (class_id & 0xff00) >> 8;
    uint16_t cls_lo = class_id & 0xff;

    if (idx < 0x100) {
%for cl_idx in range(0, len(classes_per_eng[0x6F])):
        if (cls_hi >= ${hex(class_id_to_arch_id(classes_per_eng[0x6F][cl_idx]))})
            return P_PARSE_${cl_for_function_name(classes_per_eng[0x6F][cl_idx])}_MTHD(idx);
        else
%endfor
        {
            assert(false && "unknown class id");
        }

    } else {
        switch (cls_lo) {
%for (engine_id, classes) in classes_per_eng.items():
            case ${hex(engine_id)}:
%if len(classes) == 1:
                return P_PARSE_${cl_for_function_name(classes[-1])}_MTHD(idx);
%else:
%for cl_idx in range(0, len(classes)):
                if (cls_hi >= ${hex(class_id_to_arch_id(classes[cl_idx]))})
                    return P_PARSE_${cl_for_function_name(classes[cl_idx])}_MTHD(idx);
                else
%endfor
                {
                    assert(false && "unknown class id");
                }
                break;
%endif
%endfor
            default:
                break;
        }
    }

    return "unknown method";
}

static inline void
P_DUMP_NV_MTHD_DATA(FILE *fp, uint16_t class_id, uint16_t idx, uint32_t data,
                    const char *prefix)
{
    uint16_t cls_hi = (class_id & 0xff00) >> 8;
    uint16_t cls_lo = class_id & 0xff;

    if (idx < 0x100) {
%for cl_idx in range(0, len(classes_per_eng[0x6F])):
        if (cls_hi >= ${hex(class_id_to_arch_id(classes_per_eng[0x6F][cl_idx]))})
            P_DUMP_${cl_for_function_name(classes_per_eng[0x6F][cl_idx])}_MTHD_DATA(fp, idx, data, prefix);
        else
%endfor
        {
            assert(false && "unknown class id");
        }
    } else {
        switch (cls_lo) {
%for (engine_id, classes) in classes_per_eng.items():
            case ${hex(engine_id)}:
%if len(classes) == 1:
                P_DUMP_${cl_for_function_name(classes[-1])}_MTHD_DATA(fp, idx, data, prefix);
%else:
%for cl_idx in range(0, len(classes)):
                if (cls_hi >= ${hex(class_id_to_arch_id(classes[cl_idx]))})
                    P_DUMP_${cl_for_function_name(classes[cl_idx])}_MTHD_DATA(fp, idx, data, prefix);
                else
%endfor
                {
                    assert(false && "unknown class id");
                }
%endif
                break;
%endfor
            default:
                fprintf(fp, "%s.VALUE = 0x%x\\n", prefix, data);
                break;
        }
    }
}
"""
)


def cl_for_filename(class_id):
    return f"{class_id:04x}"


def cl_for_function_name(class_id):
    return f"NV{class_id:04X}"


def class_id_to_engine_id(class_id):
    return class_id & 0xFF


def class_id_to_arch_id(class_id):
    return (class_id & 0xFF00) >> 8


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--out-h", required=False, help="Output C header.")
    parser.add_argument("--classes", help="Input class header file.", required=True)
    args = parser.parse_args()

    classes = args.classes.strip().split(" ")

    classes_per_eng = defaultdict(list)

    for cl in classes:
        class_id = int(cl.removeprefix("cl"), 16)
        engine_id = class_id_to_engine_id(class_id)

        classes_per_eng[engine_id].append(class_id)

    # Ensure everything is sorted in reverse order
    for engine_id in classes_per_eng:
        classes_per_eng[engine_id].sort(reverse=True)

    environment = {
        "classes_per_eng": classes_per_eng,
        "cl_for_filename": cl_for_filename,
        "cl_for_function_name": cl_for_function_name,
        "class_id_to_arch_id": class_id_to_arch_id,
    }

    if args.out_h is not None:
        util.write_template(args.out_h, TEMPLATE_H, environment)


if __name__ == "__main__":
    main()



# encoding=utf-8
# Copyright © 2019 Google

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import subprocess
import tempfile
import re
from collections import namedtuple


Test = namedtuple("Test", "name source match_re")

# NOTE: This test is deprecated, please add any new tests to test_gl_lower_mediump.cpp.

TESTS = [
    Test("f32 array-of-array with const index",
         """
         #version 310 es
         precision mediump float;

         uniform float in_aoa[2][2];

         layout(location = 0) out float out_color;

         void main()
         {
                 out_color = in_aoa[0][0] / in_aoa[1][1];
         }
         """,
         r'\(expression +float16_t +/'),
    Test("i32 array-of-array with const index",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform int in_aoa[2][2];

         layout(location = 0) out highp int out_color;

         void main()
         {
                 out_color = in_aoa[0][0] / in_aoa[1][1];
         }
         """,
         r'\(expression +int16_t +/'),
    Test("u32 array-of-array with const index",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform uint in_aoa[2][2];

         layout(location = 0) out highp uint out_color;

         void main()
         {
                 out_color = in_aoa[0][0] / in_aoa[1][1];
         }
         """,
         r'\(expression +uint16_t +/'),
    Test("f32 array-of-array with uniform index",
         """
         #version 310 es
         precision mediump float;

         uniform float in_aoa[2][2];
         uniform int i0, i1;

         layout(location = 0) out float out_color;

         void main()
         {
                 out_color = in_aoa[i0][i0] / in_aoa[i1][i1];
         }
         """,
         r'\(expression +float16_t +/'),
    Test("i32 array-of-array with uniform index",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform int in_aoa[2][2];
         uniform int i0, i1;

         layout(location = 0) out highp int out_color;

         void main()
         {
                 out_color = in_aoa[i0][i0] / in_aoa[i1][i1];
         }
         """,
         r'\(expression +int16_t +/'),
    Test("u32 array-of-array with uniform index",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform uint in_aoa[2][2];
         uniform int i0, i1;

         layout(location = 0) out highp uint out_color;

         void main()
         {
                 out_color = in_aoa[i0][i0] / in_aoa[i1][i1];
         }
         """,
         r'\(expression +uint16_t +/'),
    Test("f32 function",
         """
         precision mediump float;

         uniform float a, b;

         mediump float
         get_a()
         {
                 return a;
         }

         float
         get_b()
         {
                 return b;
         }

         void main()
         {
                 gl_FragColor = vec4(get_a() / get_b());
         }
         """,
         r'\(expression +float16_t +/'),
    Test("i32 function",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform int a, b;

         mediump int
         get_a()
         {
                 return a;
         }

         int
         get_b()
         {
                 return b;
         }

         out highp int color;

         void main()
         {
                 color = get_a() / get_b();
         }
         """,
         r'\(expression +int16_t +/'),
    Test("u32 function",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform uint a, b;

         mediump uint
         get_a()
         {
                 return a;
         }

         uint
         get_b()
         {
                 return b;
         }

         out highp uint color;

         void main()
         {
                 color = get_a() / get_b();
         }
         """,
         r'\(expression +uint16_t +/'),
    Test("f32 function mediump args",
         """
         precision mediump float;

         uniform float a, b;

         mediump float
         do_div(float x, float y)
         {
                 return x / y;
         }

         void main()
         {
                 gl_FragColor = vec4(do_div(a, b));
         }
         """,
         r'\(expression +float16_t +/'),
    Test("i32 function mediump args",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform int a, b;

         mediump int
         do_div(int x, int y)
         {
                 return x / y;
         }

         out highp int color;

         void main()
         {
                 color = do_div(a, b);
         }
         """,
         r'\(expression +int16_t +/'),
    Test("u32 function mediump args",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform uint a, b;

         mediump uint
         do_div(uint x, uint y)
         {
                 return x / y;
         }

         out highp uint color;

         void main()
         {
                 color = do_div(a, b);
         }
         """,
         r'\(expression +uint16_t +/'),
    Test("f32 function highp args",
         """
         precision mediump float;

         uniform float a, b;

         mediump float
         do_div(highp float x, highp float y)
         {
                 return x / y;
         }

         void main()
         {
                 gl_FragColor = vec4(do_div(a, b));
         }
         """,
         r'\(expression +float +/'),
    Test("i32 function highp args",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform int a, b;

         mediump int
         do_div(highp int x, highp int y)
         {
                 return x / y;
         }

         out highp int color;

         void main()
         {
                  color = do_div(a, b);
         }
         """,
         r'\(expression +int +/'),
    Test("u32 function highp args",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform uint a, b;

         mediump uint
         do_div(highp uint x, highp uint y)
         {
                 return x / y;
         }

         out highp uint color;

         void main()
         {
                  color = do_div(a, b);
         }
         """,
         r'\(expression +uint +/'),

    Test("f32 if",
         """
         precision mediump float;

         uniform float a, b;

         void
         main()
         {
                 if (a / b < 0.31)
                         gl_FragColor = vec4(0.0, 1.0, 0.0, 1.0);
                 else
                         gl_FragColor = vec4(1.0, 0.0, 0.0, 1.0);
         }
         """,
         r'\(expression +float16_t +/'),
    Test("i32 if",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform int a, b;

         out vec4 color;

         void
         main()
         {
                 if (a / b < 10)
                         color = vec4(0.0, 1.0, 0.0, 1.0);
                 else
                         color = vec4(1.0, 0.0, 0.0, 1.0);
         }
         """,
         r'\(expression +int16_t +/'),
    Test("u32 if",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform uint a, b;

         out vec4 color;

         void
         main()
         {
                 if (a / b < 10u)
                         color = vec4(0.0, 1.0, 0.0, 1.0);
                 else
                         color = vec4(1.0, 0.0, 0.0, 1.0);
         }
         """,
         r'\(expression +uint16_t +/'),
    Test("matrix",
         """
         precision mediump float;

         uniform vec2 a;
         uniform mat2 b;

         void main()
         {
             gl_FragColor = vec4(b * a, 0.0, 0.0);
         }
         """,
         r'\(expression f16vec2 \+ \(expression f16vec2 \* \(array_ref \(var_ref b\) \(constant int \(0\)\) \) \(swiz x \(var_ref a\) \)\) \(expression f16vec2 \* \(array_ref \(var_ref b\) \(constant int \(1\)\) \) \(swiz y \(var_ref a\) \)\) \) \)'),
    Test("f32 simple struct deref",
         """
         precision mediump float;

         struct simple {
                 float a, b;
         };

         uniform simple in_simple;

         void main()
         {
                 gl_FragColor = vec4(in_simple.a / in_simple.b);
         }
         """,
         r'\(expression +float16_t +/'),
    Test("i32 simple struct deref",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         struct simple {
                 int a, b;
         };

         uniform simple in_simple;

         out highp int color;

         void main()
         {
                 color = in_simple.a / in_simple.b;
         }
         """,
         r'\(expression +int16_t +/'),
    Test("u32 simple struct deref",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         struct simple {
                 uint a, b;
         };

         uniform simple in_simple;

         out highp uint color;

         void main()
         {
                 color = in_simple.a / in_simple.b;
         }
         """,
         r'\(expression +uint16_t +/'),
    Test("f32 embedded struct deref",
         """
         precision mediump float;

         struct simple {
                 float a, b;
         };

         struct embedded {
                 simple a, b;
         };

         uniform embedded in_embedded;

         void main()
         {
                 gl_FragColor = vec4(in_embedded.a.a / in_embedded.b.b);
         }
         """,
         r'\(expression +float16_t +/'),
    Test("i32 embedded struct deref",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         struct simple {
                 int a, b;
         };

         struct embedded {
                 simple a, b;
         };

         uniform embedded in_embedded;

         out highp int color;

         void main()
         {
                 color = in_embedded.a.a / in_embedded.b.b;
         }
         """,
         r'\(expression +int16_t +/'),
    Test("u32 embedded struct deref",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         struct simple {
                 uint a, b;
         };

         struct embedded {
                 simple a, b;
         };

         uniform embedded in_embedded;

         out highp uint color;

         void main()
         {
                 color = in_embedded.a.a / in_embedded.b.b;
         }
         """,
         r'\(expression +uint16_t +/'),
    Test("f32 arrayed struct deref",
         """
         precision mediump float;

         struct simple {
                 float a, b;
         };

         struct arrayed {
                 simple a[2];
         };

         uniform arrayed in_arrayed;

         void main()
         {
                 gl_FragColor = vec4(in_arrayed.a[0].a / in_arrayed.a[1].b);
         }
         """,
         r'\(expression +float16_t +/'),
    Test("i32 arrayed struct deref",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         struct simple {
                 int a, b;
         };

         struct arrayed {
                 simple a[2];
         };

         uniform arrayed in_arrayed;

         out highp int color;

         void main()
         {
                 color = in_arrayed.a[0].a / in_arrayed.a[1].b;
         }
         """,
         r'\(expression +int16_t +/'),
    Test("u32 arrayed struct deref",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         struct simple {
                 uint a, b;
         };

         struct arrayed {
                 simple a[2];
         };

         uniform arrayed in_arrayed;

         out highp uint color;

         void main()
         {
                 color = in_arrayed.a[0].a / in_arrayed.a[1].b;
         }
         """,
         r'\(expression +uint16_t +/'),
    Test("f32 mixed precision not lowered",
         """
         uniform mediump float a;
         uniform highp float b;

         void main()
         {
                 gl_FragColor = vec4(a / b);
         }
         """,
         r'\(expression +float +/'),
    Test("i32 mixed precision not lowered",
         """
         #version 310 es
         uniform mediump int a;
         uniform highp int b;

         out mediump int color;

         void main()
         {
                 color = a / b;
         }
         """,
         r'\(expression +int +/'),
    Test("u32 mixed precision not lowered",
         """
         #version 310 es
         uniform mediump uint a;
         uniform highp uint b;

         out mediump uint color;

         void main()
         {
                 color = a / b;
         }
         """,
         r'\(expression +uint +/'),
    Test("f32 sampler array",
         """
         #version 320 es
         precision mediump float;
         precision mediump int;

         uniform sampler2D tex[2];
         // highp shouldn't affect the return value of texture2D
         uniform highp vec2 coord;
         uniform float divisor;
         uniform int index;

         out highp vec4 color;

         void main()
         {
                 color = texture2D(tex[index], coord) / divisor;
         }
         """,
         r'\(expression +f16vec4 +/.*\(tex +f16vec4 +'),
    Test("f32 texture sample",
         """
         precision mediump float;

         uniform sampler2D tex;
         // highp shouldn't affect the return value of texture2D
         uniform highp vec2 coord;
         uniform float divisor;

         void main()
         {
                 gl_FragColor = texture2D(tex, coord) / divisor;
         }
         """,
         r'\(expression +f16vec4 +/.*\(tex +f16vec4 +'),
    Test("i32 texture sample",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform mediump isampler2D tex;
         // highp shouldn't affect the return value of texture
         uniform highp vec2 coord;
         uniform int divisor;

         out highp ivec4 color;

         void main()
         {
                 color = texture(tex, coord) / divisor;
         }
         """,
         r'\(expression +i16vec4 +/.*\(tex +i16vec4 +'),
    Test("u32 texture sample",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform mediump usampler2D tex;
         // highp shouldn't affect the return value of texture
         uniform highp vec2 coord;
         uniform uint divisor;

         out highp uvec4 color;

         void main()
         {
                 color = texture(tex, coord) / divisor;
         }
         """,
         r'\(expression +u16vec4 +/.*\(tex +u16vec4 +'),
    Test("f32 image array",
         """
         #version 320 es
         precision mediump float;

         layout(rgba16f) readonly uniform mediump image2D img[2];
         // highp shouldn't affect the return value of imageLoad
         uniform highp ivec2 coord;
         uniform float divisor;

         out highp vec4 color;

         void main()
         {
                 color = imageLoad(img[1], coord) / divisor;
         }
         """,
         r'\(expression +f16vec4 +/'),
    Test("f32 image load",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         layout(rgba16f) readonly uniform mediump image2D img;
         // highp shouldn't affect the return value of imageLoad
         uniform highp ivec2 coord;
         uniform float divisor;

         out highp vec4 color;

         void main()
         {
                 color = imageLoad(img, coord) / divisor;
         }
         """,
         r'\(expression +f16vec4 +/'),
    Test("i32 image load",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         layout(rgba16i) readonly uniform mediump iimage2D img;
         // highp shouldn't affect the return value of imageLoad
         uniform highp ivec2 coord;
         uniform int divisor;

         out highp ivec4 color;

         void main()
         {
                 color = imageLoad(img, coord) / divisor;
         }
         """,
         r'\(expression +i16vec4 +/'),
    Test("u32 image load",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         layout(rgba16ui) readonly uniform mediump uimage2D img;
         // highp shouldn't affect the return value of imageLoad
         uniform highp ivec2 coord;
         uniform uint divisor;

         out highp uvec4 color;

         void main()
         {
                 color = imageLoad(img, coord) / divisor;
         }
         """,
         r'\(expression +u16vec4 +/'),
    Test("f32 expression in lvalue",
         """
         uniform mediump float a, b;

         void main()
         {
                 gl_FragColor = vec4(1.0);
                 gl_FragColor[int(a / b)] = 0.5;
         }
         """,
         r'\(expression +float16_t +/'),
    Test("i32 expression in lvalue",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform mediump int a, b;

         out vec4 color;

         void main()
         {
                 color = vec4(1.0);
                 color[a / b] = 0.5;
         }
         """,
         r'\(expression +int16_t +/'),
    Test("f32 builtin with const arg",
         """
         uniform mediump float a;

         void main()
         {
                 gl_FragColor = vec4(min(a, 3.0));
         }
         """,
         r'\(expression +float16_t min'),
    Test("i32 builtin with const arg",
         """
         #version 310 es
         uniform mediump int a;

         out highp int color;

         void main()
         {
                 color = min(a, 3);
         }
         """,
         r'\(expression +int16_t min'),
    Test("u32 builtin with const arg",
         """
         #version 310 es
         uniform mediump uint a;

         out highp uint color;

         void main()
         {
                 color = min(a, 3u);
         }
         """,
         r'\(expression +uint16_t min'),
    Test("dFdx",
         """
         #version 300 es
         precision mediump float;

         in vec4 var;
         out vec4 color;

         void main()
         {
                 color = dFdx(var);
         }
         """,
         r'\(expression +f16vec4 +dFdx +\(expression +f16vec4'),
    Test("dFdy",
         """
         #version 300 es
         precision mediump float;

         in vec4 var;
         out vec4 color;

         void main()
         {
                 color = dFdy(var);
         }
         """,
         r'\(expression +f16vec4 +dFdy +\(expression +f16vec4'),
    Test("textureSize",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform mediump sampler2D tex;
         out ivec2 color;

         void main()
         {
                 color = textureSize(tex, 0) * ivec2(2);
         }
         """,
         r'expression ivec2 \* \(txs ivec2 \(var_ref tex'),
    Test("floatBitsToInt",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform float val;
         out int color;

         void main()
         {
                 color = floatBitsToInt(val + 1.0) + 1;
         }
         """,
         r'expression int bitcast_f2i \(expression float'),
    Test("floatBitsToUint",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform float val;
         out uint color;

         void main()
         {
                 color = floatBitsToUint(val + 1.0) + 1u;
         }
         """,
         r'expression uint bitcast_f2u \(expression float'),
    Test("intBitsToFloat",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform int val;
         out float color;

         void main()
         {
                 color = intBitsToFloat(val + 1) + 1.0;
         }
         """,
         r'expression float bitcast_i2f \(expression int'),
    Test("uintBitsToFloat",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform uint val;
         out float color;

         void main()
         {
                 color = uintBitsToFloat(val + 1u) + 1.0;
         }
         """,
         r'expression float bitcast_u2f \(expression uint'),
    Test("bitfieldReverse",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform int val;
         out int color;

         void main()
         {
                 color = bitfieldReverse(val + 1) + 1;
         }
         """,
         r'expression int bitfield_reverse \(expression int'),
    Test("frexp",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform float val;
         out float color;
         out int color2;

         void main()
         {
                 int y;
                 float x = frexp(val + 1.0, y);
                 color = x + 1.0;
                 color2 = y + 1;
         }
         """,
         r'expression int16_t i2imp \(expression int frexp_exp \(expression float f162f'),
    Test("ldexp",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform float val;
         uniform int exp;
         out float color;

         void main()
         {
                 color = ldexp(val + 1.0, exp + 1) + 1.0;
         }
         """,
         r'expression float ldexp \(expression float'),
    Test("uaddCarry",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform uint x, y;
         out uint color;

         void main()
         {
                 lowp uint carry;
                 color = uaddCarry(x * 2u, y * 2u, carry) * 2u;
                 color *= carry;
         }
         """,
         r'expression uint \+ \(expression uint u2u \(expression uint16_t \* \(expression uint16_t u2ump \(var_ref x\) \) \(constant uint16_t \(2\)\) \) \) \(expression uint u2u \(expression uint16_t \* \(expression uint16_t u2ump \(var_ref y'),
    Test("usubBorrow",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform uint x, y;
         out uint color;

         void main()
         {
                 lowp uint borrow;
                 color = usubBorrow(x * 2u, y * 2u, borrow) * 2u;
                 color *= borrow;
         }
         """,
         r'expression uint \- \(expression uint u2u \(expression uint16_t \* \(expression uint16_t u2ump \(var_ref x\) \) \(constant uint16_t \(2\)\) \) \) \(expression uint u2u \(expression uint16_t \* \(expression uint16_t u2ump \(var_ref y'),
    Test("imulExtended",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform int x, y;
         out int color;

         void main()
         {
                 int msb, lsb;
                 imulExtended(x + 2, y + 2, msb, lsb);
                 color = msb + lsb;
         }
         """,
         r'expression int64_t \* \(expression int'),
    Test("umulExtended",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform uint x, y;
         out uint color;

         void main()
         {
                 uint msb, lsb;
                 umulExtended(x + 2u, y + 2u, msb, lsb);
                 color = msb + lsb;
         }
         """,
         r'expression uint64_t \* \(expression uint'),
    Test("unpackUnorm2x16",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform uint val;
         out vec2 color;

         void main()
         {
                 color = unpackUnorm2x16(val + 1u) + vec2(1.0);
         }
         """,
         r'expression vec2 unpackUnorm2x16 \(expression uint'),
    Test("unpackSnorm2x16",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform uint val;
         out vec2 color;

         void main()
         {
                 color = unpackSnorm2x16(val + 1u) + vec2(1.0);
         }
         """,
         r'expression vec2 unpackSnorm2x16 \(expression uint'),
    Test("packUnorm2x16",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform vec2 val;
         out uint color;

         void main()
         {
                 color = packUnorm2x16(val + vec2(1.0)) + 1u;
         }
         """,
         r'expression uint packUnorm2x16 \(expression vec2'),
    Test("packSnorm2x16",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform vec2 val;
         out uint color;

         void main()
         {
                 color = packSnorm2x16(val + vec2(1.0)) + 1u;
         }
         """,
         r'expression uint packSnorm2x16 \(expression vec2'),
    Test("packHalf2x16",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform vec2 val;
         out uint color;

         void main()
         {
                 color = packHalf2x16(val + vec2(1.0)) + 1u;
         }
         """,
         r'expression uint packHalf2x16 \(expression vec2'),
    Test("packUnorm4x8",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform vec4 val;
         out uint color;

         void main()
         {
                 color = packUnorm4x8(val + vec4(1.0)) + 1u;
         }
         """,
         r'expression uint packUnorm4x8 \(expression vec4'),
    Test("packSnorm4x8",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform vec4 val;
         out uint color;

         void main()
         {
                 color = packSnorm4x8(val + vec4(1.0)) + 1u;
         }
         """,
         r'expression uint packSnorm4x8 \(expression vec4'),
    Test("interpolateAtCentroid",
         """
         #version 320 es
         precision mediump float;
         precision mediump int;

         in float val;
         out float color;

         void main()
         {
                 color = interpolateAtCentroid(val) + 1.0;
         }
         """,
         r'expression float16_t interpolate_at_centroid \(expression float16_t'),
    Test("interpolateAtOffset",
         """
         #version 320 es
         precision mediump float;
         precision mediump int;

         uniform highp vec2 offset;
         in float val;
         out float color;

         void main()
         {
                 color = interpolateAtOffset(val, offset) + 1.0;
         }
         """,
         r'expression float16_t interpolate_at_offset \(expression float16_t'),
    Test("interpolateAtSample",
         """
         #version 320 es
         precision mediump float;
         precision mediump int;

         uniform highp int sample_index;
         in float val;
         out float color;

         void main()
         {
                 color = interpolateAtSample(val, sample_index) + 1.0;
         }
         """,
         r'expression float16_t interpolate_at_sample \(expression float16_t'),
    Test("bitfieldExtract",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform highp int offset, bits;
         uniform int val;
         out int color;

         void main()
         {
                 color = bitfieldExtract(val, offset, bits) + 1;
         }
         """,
         r'expression int16_t bitfield_extract \(expression int16_t'),
    Test("bitfieldInsert",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform highp int offset, bits;
         uniform int val, val2;
         out int color;

         void main()
         {
                 color = bitfieldInsert(val, val2, offset, bits) + 1;
         }
         """,
         r'expression int16_t bitfield_insert \(expression int16_t'),
    Test("bitCount",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform highp int val;
         out int color;

         void main()
         {
                 color = bitCount(val) + 1;
         }
         """,
         r'expression int16_t \+ \(expression int16_t i2imp \(expression int bit_count \(var_ref val'),
    Test("findLSB",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform highp int val;
         out int color;

         void main()
         {
                 color = findLSB(val) + 1;
         }
         """,
         r'expression int16_t \+ \(expression int16_t i2imp \(expression int find_lsb \(var_ref val'),
    Test("findMSB",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform highp int val;
         out int color;

         void main()
         {
                 color = findMSB(val) + 1;
         }
         """,
         r'expression int16_t \+ \(expression int16_t i2imp \(expression int find_msb \(var_ref val'),
    Test("unpackHalf2x16",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform highp uint val;
         out vec2 color;

         void main()
         {
                 color = unpackHalf2x16(val) + vec2(1.0);
         }
         """,
         r'expression f16vec2 \+ \(expression f16vec2 f2fmp \(expression vec2 unpackHalf2x16 \(var_ref val'),
    Test("unpackUnorm4x8",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform highp uint val;
         out vec4 color;

         void main()
         {
                 color = unpackUnorm4x8(val) + vec4(1.0);
         }
         """,
         r'expression f16vec4 \+ \(expression f16vec4 f2fmp \(expression vec4 unpackUnorm4x8 \(var_ref val'),
    Test("unpackSnorm4x8",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform highp uint val;
         out vec4 color;

         void main()
         {
                 color = unpackSnorm4x8(val) + vec4(1.0);
         }
         """,
         r'expression f16vec4 \+ \(expression f16vec4 f2fmp \(expression vec4 unpackSnorm4x8 \(var_ref val'),
    Test("f32 csel",
         """
         #version 300 es
         precision mediump float;

         in vec4 var;
         out vec4 color;

         void main()
         {
                 color = (var.x > var.y) ? var : vec4(10.0);
         }
         """,
         r'\(expression f16vec4 f2fmp \(constant vec4 \(10'),
    Test("i32 csel",
         """
         #version 310 es
         precision mediump int;

         in flat ivec4 var;
         out ivec4 color;

         void main()
         {
                 color = (var.x > var.y) ? var : ivec4(10);
         }
         """,
         r'\(expression i16vec4 i2imp \(constant ivec4 \(10'),
    Test("u32 csel",
         """
         #version 310 es
         precision mediump int;

         in flat uvec4 var;
         out uvec4 color;

         void main()
         {
                 color = (var.x > var.y) ? var : uvec4(10);
         }
         """,
         r'\(expression u16vec4 u2ump \(constant uvec4 \(10'),
    Test("f32 loop counter",
         """
         #version 300 es
         precision mediump float;

         uniform float n, incr;
         out float color;

         void main()
         {
                 color = 0.0;
                 for (float x = 0.0; x < n; x += incr)
                    color += x;
         }
         """,
         r'\(assign  \(x\) \(var_ref x\)  \(expression float16_t \+ \(var_ref x\) \(var_ref incr'),
    Test("i32 loop counter",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform int n, incr;
         out int color;

         void main()
         {
                 color = 0;
                 for (int x = 0; x < n; x += incr)
                    color += x;
         }
         """,
         r'\(assign  \(x\) \(var_ref x\)  \(expression int16_t \+ \(var_ref x\) \(expression int16_t i2imp \(var_ref incr'),
    Test("u32 loop counter",
         """
         #version 310 es
         precision mediump float;
         precision mediump int;

         uniform uint n, incr;
         out uint color;

         void main()
         {
                 color = 0u;
                 for (uint x = 0u; x < n; x += incr)
                    color += x;
         }
         """,
         r'\(assign  \(x\) \(var_ref x\)  \(expression uint16_t \+ \(var_ref x\) \(expression uint16_t u2ump \(var_ref incr'),
    Test("f32 temp array",
         """
         #version 300 es
         precision mediump float;

         uniform float x,y;
         out float color;

         void main()
         {
                 float a[2] = float[2](x, y);
                 if (x > 0.0)
                     a[1] = 3.0;
                 color = a[0] + a[1];
         }
         """,
         r'\(expression float16_t f2fmp \(constant float \(3'),
    Test("i32 temp array",
         """
         #version 310 es
         precision mediump int;

         uniform int x,y;
         out int color;

         void main()
         {
                 int a[2] = int[2](x, y);
                 if (x > 0)
                     a[1] = 3;
                 color = a[0] + a[1];
         }
         """,
         r'\(expression int16_t i2imp \(constant int \(3'),
    Test("u32 temp array",
         """
         #version 310 es
         precision mediump int;

         uniform uint x,y;
         out uint color;

         void main()
         {
                 uint a[2] = uint[2](x, y);
                 if (x > 0u)
                     a[1] = 3u;
                 color = a[0] + a[1];
         }
         """,
         r'\(expression uint16_t u2ump \(constant uint \(3'),
    Test("f32 temp array of array",
         """
         #version 310 es
         precision mediump float;

         uniform float x,y;
         out float color;

         void main()
         {
                 float a[2][2] = float[2][2](float[2](x, y), float[2](x, y));
                 if (x > 0.0)
                     a[1][1] = 3.0;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression float16_t f2fmp \(constant float \(3'),
    Test("i32 temp array of array",
         """
         #version 310 es
         precision mediump int;

         uniform int x,y;
         out int color;

         void main()
         {
                 int a[2][2] = int[2][2](int[2](x, y), int[2](x, y));
                 if (x > 0)
                     a[1][1] = 3;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression int16_t i2imp \(constant int \(3'),
    Test("u32 temp array of array",
         """
         #version 310 es
         precision mediump int;

         uniform uint x,y;
         out uint color;

         void main()
         {
                 uint a[2][2] = uint[2][2](uint[2](x, y), uint[2](x, y));
                 if (x > 0u)
                     a[1][1] = 3u;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression uint16_t u2ump \(constant uint \(3'),
    Test("f32 temp array of array assigned from highp",
         """
         #version 310 es
         precision mediump float;

         uniform float x,y;
         out float color;

         void main()
         {
                 highp float b[2][2] = float[2][2](float[2](x, y), float[2](x, y));
                 float a[2][2];
                 a = b;
                 if (x > 0.0)
                     a[1][1] = 3.0;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression float16_t f2fmp \(constant float \(3'),
    Test("i32 temp array of array assigned from highp",
         """
         #version 310 es
         precision mediump int;

         uniform int x,y;
         out int color;

         void main()
         {
                 highp int b[2][2] = int[2][2](int[2](x, y), int[2](x, y));
                 int a[2][2];
                 a = b;
                 if (x > 0)
                     a[1][1] = 3;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression int16_t i2imp \(constant int \(3'),
    Test("u32 temp array of array assigned from highp",
         """
         #version 310 es
         precision mediump int;

         uniform uint x,y;
         out uint color;

         void main()
         {
                 highp uint b[2][2] = uint[2][2](uint[2](x, y), uint[2](x, y));
                 uint a[2][2];
                 a = b;
                 if (x > 0u)
                     a[1][1] = 3u;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression uint16_t u2ump \(constant uint \(3'),
    Test("f32 temp array of array assigned to highp",
         """
         #version 310 es
         precision mediump float;

         uniform float x,y;
         out float color;

         void main()
         {
                 float a[2][2] = float[2][2](float[2](x, y), float[2](x, y));
                 highp float b[2][2];
                 b = a;
                 a = b;
                 if (x > 0.0)
                     a[1][1] = 3.0;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression float16_t f2fmp \(constant float \(3'),
    Test("i32 temp array of array assigned to highp",
         """
         #version 310 es
         precision mediump int;

         uniform int x,y;
         out int color;

         void main()
         {
                 int a[2][2] = int[2][2](int[2](x, y), int[2](x, y));
                 highp int b[2][2];
                 b = a;
                 a = b;
                 if (x > 0)
                     a[1][1] = 3;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression int16_t i2imp \(constant int \(3'),
    Test("u32 temp array of array assigned to highp",
         """
         #version 310 es
         precision mediump int;

         uniform uint x,y;
         out uint color;

         void main()
         {
                 uint a[2][2] = uint[2][2](uint[2](x, y), uint[2](x, y));
                 highp uint b[2][2];
                 b = a;
                 a = b;
                 if (x > 0u)
                     a[1][1] = 3u;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression uint16_t u2ump \(constant uint \(3'),
    Test("f32 temp array of array returned by function",
         """
         #version 310 es
         precision mediump float;

         uniform float x,y;
         out float color;

         float[2][2] f(void)
         {
            return float[2][2](float[2](x, y), float[2](x, y));
         }

         void main()
         {
                 float a[2][2] = f();
                 if (x > 0.0)
                     a[1][1] = 3.0;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression float16_t f2fmp \(constant float \(3'),
    Test("i32 temp array of array returned by function",
         """
         #version 310 es
         precision mediump int;

         uniform int x,y;
         out int color;

         int[2][2] f(void)
         {
            return int[2][2](int[2](x, y), int[2](x, y));
         }

         void main()
         {
                 int a[2][2] = f();
                 if (x > 0)
                     a[1][1] = 3;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression int16_t i2imp \(constant int \(3'),
    Test("u32 temp array of array returned by function",
         """
         #version 310 es
         precision mediump int;

         uniform uint x,y;
         out uint color;

         uint[2][2] f(void)
         {
            return uint[2][2](uint[2](x, y), uint[2](x, y));
         }

         void main()
         {
                 uint a[2][2] = f();
                 if (x > 0u)
                     a[1][1] = 3u;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression uint16_t u2ump \(constant uint \(3'),
    Test("f32 temp array of array as function out",
         """
         #version 310 es
         precision mediump float;

         uniform float x,y;
         out float color;

         void f(out float[2][2] v)
         {
            v = float[2][2](float[2](x, y), float[2](x, y));
         }

         void main()
         {
                 float a[2][2];
                 f(a);
                 if (x > 0.0)
                     a[1][1] = 3.0;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression float16_t f2fmp \(constant float \(3'),
    Test("i32 temp array of array as function out",
         """
         #version 310 es
         precision mediump int;

         uniform int x,y;
         out int color;

         void f(out int[2][2] v)
         {
            v = int[2][2](int[2](x, y), int[2](x, y));
         }

         void main()
         {
                 int a[2][2];
                 f(a);
                 if (x > 0)
                     a[1][1] = 3;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression int16_t i2imp \(constant int \(3'),
    Test("u32 temp array of array as function out",
         """
         #version 310 es
         precision mediump int;

         uniform uint x,y;
         out uint color;

         void f(out uint[2][2] v)
         {
            v = uint[2][2](uint[2](x, y), uint[2](x, y));
         }

         void main()
         {
                 uint a[2][2];
                 f(a);
                 if (x > 0u)
                     a[1][1] = 3u;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression uint16_t u2ump \(constant uint \(3'),
    Test("f32 temp array of array as function in",
         """
         #version 310 es
         precision mediump float;

         uniform float x,y;
         out float color;

         float[2][2] f(in float[2][2] v)
         {
            float t[2][2] = v;
            return t;
         }

         void main()
         {
                 float a[2][2];
                 a = f(a);
                 if (x > 0.0)
                     a[1][1] = 3.0;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression float16_t f2fmp \(constant float \(3'),
    Test("i32 temp array of array as function in",
         """
         #version 310 es
         precision mediump int;

         uniform int x,y;
         out int color;

         int[2][2] f(in int[2][2] v)
         {
            int t[2][2] = v;
            return t;
         }

         void main()
         {
                 int a[2][2];
                 a = f(a);
                 if (x > 0)
                     a[1][1] = 3;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression int16_t i2imp \(constant int \(3'),
    Test("u32 temp array of array as function in",
         """
         #version 310 es
         precision mediump int;

         uniform uint x,y;
         out uint color;

         uint[2][2] f(in uint[2][2] v)
         {
            uint t[2][2] = v;
            return t;
         }

         void main()
         {
                 uint a[2][2];
                 a = f(a);
                 if (x > 0u)
                     a[1][1] = 3u;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression uint16_t u2ump \(constant uint \(3'),
    Test("f32 temp array of array as function inout",
         """
         #version 310 es
         precision mediump float;

         uniform float x,y;
         out float color;

         void f(inout float[2][2] v)
         {
            float t[2][2] = v;
            v = t;
         }

         void main()
         {
                 float a[2][2];
                 f(a);
                 if (x > 0.0)
                     a[1][1] = 3.0;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression float16_t f2fmp \(constant float \(3'),
    Test("i32 temp array of array as function inout",
         """
         #version 310 es
         precision mediump int;

         uniform int x,y;
         out int color;

         void f(inout int[2][2] v)
         {
            int t[2][2] = v;
            v = t;
         }

         void main()
         {
                 int a[2][2];
                 f(a);
                 if (x > 0)
                     a[1][1] = 3;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression int16_t i2imp \(constant int \(3'),
    Test("u32 temp array of array as function inout",
         """
         #version 310 es
         precision mediump int;

         uniform uint x,y;
         out uint color;

         void f(inout uint[2][2] v)
         {
            uint t[2][2] = v;
            v = t;
         }

         void main()
         {
                 uint a[2][2];
                 f(a);
                 if (x > 0u)
                     a[1][1] = 3u;
                 color = a[0][0] + a[1][1];
         }
         """,
         r'\(expression uint16_t u2ump \(constant uint \(3'),
    Test("f32 temp struct (not lowered in the presence of control flow - TODO)",
         """
         #version 300 es
         precision mediump float;

         uniform float x,y;
         out float color;

         void main()
         {
                 struct { float x,y; } s;
                 s.x = x;
                 s.y = y;
                 if (x > 0.0)
                     s.y = 3.0;
                 color = s.x + s.y;
         }
         """,
         r'\(constant float \(3'), # should be float16_t
    Test("i32 temp struct (not lowered in the presence of control flow - TODO)",
         """
         #version 300 es
         precision mediump int;

         uniform int x,y;
         out int color;

         void main()
         {
                 struct { int x,y; } s;
                 s.x = x;
                 s.y = y;
                 if (x > 0)
                     s.y = 3;
                 color = s.x + s.y;
         }
         """,
         r'\(constant int \(3'), # should be int16_t
    Test("u32 temp struct (not lowered in the presence of control flow - TODO)",
         """
         #version 300 es
         precision mediump int;

         uniform uint x,y;
         out uint color;

         void main()
         {
                 struct { uint x,y; } s;
                 s.x = x;
                 s.y = y;
                 if (x > 0u)
                     s.y = 3u;
                 color = s.x + s.y;
         }
         """,
         r'\(constant uint \(3'), # should be uint16_t

    Test("vec4 constructor from float",
         """
         uniform highp float a;
         uniform mediump float b;

         void main()
         {
                 gl_FragColor = vec4(a) * b;
         }
         """,
         r'\(expression vec4 \* \(swiz xxxx \(var_ref a\) \)\(expression float f162f \(var_ref b\) \) \)'),

    Test("respect copies",
         """
         uniform mediump float a, b;

         void main()
         {
            highp float x = a;
            gl_FragColor.x = x * b;
         }
         """,
         r'expression float \* \(expression float f162f \(var_ref a\) \) \(expression float f162f \(var_ref b\) \) '), # should be uint16_t

    Test("conversion constructor precision",
         """
         #version 300 es
         uniform mediump uint a;
         out highp float result;

         void main()
         {
            /* Constructors don't have a precision qualifier themselves, but
             * constructors are an operation, and so they do the usual "get
             * precision from my operands, or default to the precision of the
             * lvalue" rule.  So, the u2f is done at mediump due to a's precision.
             */
            result = float(a);
         }
         """,
         r'expression float16_t u2f \(expression uint16_t u2ump \(var_ref a\) \)'), # should be uint16_t

]


def compile_shader(standalone_compiler, source):
    with tempfile.NamedTemporaryFile(mode='wt', suffix='.frag') as source_file:
        print(source, file=source_file)
        source_file.flush()
        return subprocess.check_output([standalone_compiler,
                                        '--version', '300',
                                        '--lower-precision',
                                        '--dump-lir',
                                        source_file.name],
                                       universal_newlines=True)


def run_test(standalone_compiler, test):
    ir = compile_shader(standalone_compiler, test.source)

    if re.search(test.match_re, ir) is None:
        print(ir)
        return False

    return True


def main():
    standalone_compiler = sys.argv[1]
    passed = 0

    for test in TESTS:
        print('Testing {} ... '.format(test.name), end='')

        result = run_test(standalone_compiler, test)

        if result:
            print('PASS')
            passed += 1
        else:
            print('FAIL')

    print('{}/{} tests returned correct results'.format(passed, len(TESTS)))
    sys.exit(0 if passed == len(TESTS) else 1)


if __name__ == '__main__':
    main()



# Copyright © 2025 Imagination Technologies Ltd.
# SPDX-License-Identifier: MIT

import argparse
import sys
import math

a = 'a'
b = 'b'

lower_algebraic = []
lower_algebraic_late = []

def lowered_fround_even(src):
   abs_src = ('fabs', src)
   ffloor_temp = ('ffloor', abs_src)
   ffract_temp = ('ffract', abs_src)

   ceil_temp = ('fadd', ffloor_temp, 1.0)
   even_temp = ('fmul', ffloor_temp, 0.5)

   even_ffract_temp = ('ffract', even_temp)

   ishalf_temp = ('feq', ffract_temp, 0.5)
   ffract_temp = ('bcsel', ishalf_temp, even_ffract_temp, ffract_temp)

   lesshalf_temp = ('flt', ffract_temp, 0.5)
   result_temp = ('bcsel', lesshalf_temp, ffloor_temp, ceil_temp)

   res = ('fcopysign_pco', result_temp, src)
   return res

lower_algebraic.append((('fround_even', a), lowered_fround_even(a)))

lower_insert_extract = [
   (('insert_u8', 'a@32', b), ('bitfield_insert', 0, a, ('imul', b, 8), 8)),
   (('insert_u16', 'a@32', b), ('bitfield_insert', 0, a, ('imul', b, 16), 16)),

   (('extract_u8', 'a@32', b), ('ubitfield_extract', a, ('imul', b, 8), 8)),
   (('extract_i8', 'a@32', b), ('ibitfield_extract', a, ('imul', b, 8), 8)),

   (('extract_u16', 'a@32', b), ('ubitfield_extract', a, ('imul', b, 16), 16)),
   (('extract_i16', 'a@32', b), ('ibitfield_extract', a, ('imul', b, 16), 16)),
]

lower_algebraic_late.extend(lower_insert_extract)

lower_b2b = [
   (('b2b32', a), ('ineg', ('b2i32', a))),
   (('b2b1', a), ('ine', a, 0)),
]

lower_algebraic.extend(lower_b2b)

lower_scmp = [
   # Float comparisons + bool conversions.
   (('b2f32', ('flt', a, b)), ('slt', a, 'b@32'), '!options->lower_scmp'),
   (('b2f32', ('fge', a, b)), ('sge', a, 'b@32'), '!options->lower_scmp'),
   (('b2f32', ('feq', a, b)), ('seq', a, 'b@32'), '!options->lower_scmp'),
   (('b2f32', ('fneu', a, b)), ('sne', a, 'b@32'), '!options->lower_scmp'),

   # Float comparisons + bool conversions via bcsel.
   (('bcsel@32', ('flt', a, b), 1.0, 0.0), ('slt', a, 'b@32'), '!options->lower_scmp'),
   (('bcsel@32', ('fge', a, b), 1.0, 0.0), ('sge', a, 'b@32'), '!options->lower_scmp'),
   (('bcsel@32', ('feq', a, b), 1.0, 0.0), ('seq', a, 'b@32'), '!options->lower_scmp'),
   (('bcsel@32', ('fneu', a, b), 1.0, 0.0), ('sne', a, 'b@32'), '!options->lower_scmp'),
]

lower_algebraic_late.extend(lower_scmp)

# TODO: core-specific info.
params=[]

def run():
    import nir_algebraic  # pylint: disable=import-error
    print('#include "pco_internal.h"')
    print(nir_algebraic.AlgebraicPass('pco_nir_lower_algebraic', lower_algebraic, params=params).render())
    print(nir_algebraic.AlgebraicPass('pco_nir_lower_algebraic_late', lower_algebraic_late, params=params).render())

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--import-path', required=True)
    args = parser.parse_args()
    sys.path.insert(0, args.import_path)
    run()

if __name__ == '__main__':
    main()



# Copyright (C) 2020 Collabora, Ltd.
# SPDX-License-Identifier: MIT

TEMPLATE = """#include "bi_opcodes.h"
<%
def hasmod(mods, name):
        return 1 if name in mods else 0
%>
struct bi_op_props bi_opcode_props[BI_NUM_OPCODES] = {
% for opcode in sorted(mnemonics):
    <%
        add = instructions["+" + opcode][0][1] if "+" + opcode in instructions else None
        size = typesize(opcode)
        message = add["message"].upper() if add else "NONE"
        sr_count = add["staging_count"].upper() if add else "0"
        sr_read = int(add["staging"] in ["r", "rw"] if add else False)
        sr_write = int(add["staging"] in ["w", "rw"] if add else False)
        last = int(bool(add["last"]) if add else False)
        table = int(bool(add["table"]) if add else False)
        branch = int(opcode.startswith('BRANCH'))
        has_fma = int("*" + opcode in instructions)
        has_add = int("+" + opcode in instructions)
        is_float = int(opcode.endswith("f32") or opcode.endswith("f16"))
        mods = ops[opcode]['modifiers']
        clamp = hasmod(mods, 'clamp')
        not_result = hasmod(mods, 'not_result')
        abs = hasmod(mods, 'abs0') | (hasmod(mods, 'abs1') << 1) | (hasmod(mods, 'abs2') << 2)
        neg = hasmod(mods, 'neg0') | (hasmod(mods, 'neg1') << 1) | (hasmod(mods, 'neg2') << 2)
        m_not = hasmod(mods, 'not1')
    %>
    [BI_OPCODE_${opcode.replace('.', '_').upper()}] = {
        "${opcode}", BIFROST_MESSAGE_${message}, BI_SIZE_${size},
        BI_SR_COUNT_${sr_count}, ${sr_read}, ${sr_write}, ${last}, ${branch},
        ${table}, ${has_fma}, ${has_add}, ${is_float}, ${clamp}, ${not_result},
        ${abs}, ${neg}, ${m_not},
    },
% endfor
};"""

import sys
from bifrost_isa import *
from mako.template import Template

instructions = {}
for arg in sys.argv[1:]:
    new_instructions = parse_instructions(arg, include_pseudo = True)
    instructions.update(new_instructions)

ir_instructions = partition_mnemonics(instructions)
mnemonics = set(x[1:] for x in instructions.keys())

print(Template(COPYRIGHT + TEMPLATE).render(ops = ir_instructions, mnemonics = mnemonics, instructions = instructions, typesize = typesize))



#!/usr/bin/env python3
# coding=utf-8
##########################################################################
#
# enums2names - Parse and convert enums to translator code
# (C) Copyright 2021 Matti 'ccr' Hämäläinen <ccr@tnsp.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sub license, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice (including the
# next paragraph) shall be included in all copies or substantial portions
# of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
# IN NO EVENT SHALL VMWARE AND/OR ITS SUPPLIERS BE LIABLE FOR
# ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
##########################################################################

import sys
import os.path
import re
import signal
import argparse
import textwrap

assert sys.version_info >= (3, 6)


#
# List of enums we wish to include in output.
# NOTE: This needs to be updated if such enums are added.
#
lst_enum_include = [
    "pipe_texture_target",
    "pipe_shader_ir",
    "pipe_map_flags",
    "pipe_video_cap",
    "pipe_video_profile",
    "pipe_video_entrypoint",
    "pipe_video_vpp_orientation",
    "pipe_video_vpp_blend_mode",
    "pipe_resource_param",
    "pipe_fd_type",
    "pipe_blendfactor",
    "pipe_blend_func",
    "pipe_logicop",
]


###
### Utility functions
###
## Fatal error handler
def pkk_fatal(smsg):
    print("ERROR: "+ smsg)
    sys.exit(1)


## Handler for SIGINT signals
def pkk_signal_handler(signal, frame):
    print("\nQuitting due to SIGINT / Ctrl+C!")
    sys.exit(1)


## Argument parser subclass
class PKKArgumentParser(argparse.ArgumentParser):
    def print_help(self):
        print("enums2names - Parse and convert enums to translator code\n"
        "(C) Copyright 2021 Matti 'ccr' Hämäläinen <ccr@tnsp.org>\n")
        super().print_help()

    def error(self, msg):
        self.print_help()
        print(f"\nERROR: {msg}", file=sys.stderr)
        sys.exit(2)


def pkk_get_argparser():
    optparser = PKKArgumentParser(
        usage="%(prog)s [options] <infile|->\n"
        "example: %(prog)s ../../include/pipe/p_defines.h -C tr_util.c -H tr_util.h"
        )

    optparser.add_argument("in_files",
        type=str,
        metavar="infiles",
        nargs="+",
        help="path to input header files (or '-' for stdin)")

    optparser.add_argument("-C",
        type=str,
        metavar="outfile",
        dest="out_source",
        help="output C source file")

    optparser.add_argument("-H",
        type=str,
        metavar="outfile",
        dest="out_header",
        help="output H header file (the basename will be use to include the header in the C source file)")

    return optparser


class PKKHeaderParser:

    def __init__(self, nfilename):
        self.filename = nfilename
        self.enums = {}
        self.state = 0
        self.nline = 0
        self.mdata = []
        self.start = 0
        self.name = None
        self.in_multiline_comment = False

    def error(self, msg):
        pkk_fatal(f"{self.filename}:{self.nline} : {msg}")

    def parse_line(self, sline: str):
        start = sline.find('/*')
        end = sline.find('*/')
        if not self.in_multiline_comment and start >= 0:
            if end >= 0:
                assert end > start
                sline = sline[:start] + sline[end + 2:]
            else:
                sline = sline[:start]
                self.in_multiline_comment = True
        elif self.in_multiline_comment and end >= 0:
            self.in_multiline_comment = False
            sline = sline[end + 2:]
        elif self.in_multiline_comment:
            return
        # A kingdom for Py3.8 := operator ...
        smatch = re.match(r'^enum\s+([A-Za-z0-9_]+)\s+.*;', sline)
        if smatch:
            pass
        else:
            smatch = re.match(r'^enum\s+([A-Za-z0-9_]+)', sline)
            if smatch:
                stmp = smatch.group(1)

                if self.state != 0:
                    self.error(f"enum '{stmp}' starting inside another enum '{self.name}'")

                self.name = stmp
                self.state = 1
                self.start = self.nline
                self.mdata = []
            else:
                smatch = re.match(r'^}(\s*|\s*[A-Z][A-Z_]+\s*);', sline)
                if smatch:
                    if self.state == 1:
                        if self.name in self.enums:
                            self.error("duplicate enum definition '{}', lines {} - {} vs {} - {}".format(
                            self.name, self.enums[self.name]["start"], self.enums[self.name]["end"],
                            self.start, self.nline))

                        self.enums[self.name] = {
                            "data": self.mdata,
                            "start": self.start,
                            "end": self.nline
                        }

                    self.state = 0

                elif self.state == 1:
                    smatch = re.match(r'([A-Za-z0-9_]+)\s*=\s*(.+)\s*,?', sline)
                    if smatch:
                        self.mdata.append(smatch.group(1))
                    else:
                        smatch = re.match(r'([A-Za-z0-9_]+)\s*,?', sline)
                        if smatch:
                            self.mdata.append(smatch.group(1))

    def parse_file(self, fh):
        self.nline = 0
        for line in fh:
            self.nline += 1
            self.parse_line(line.strip())

        return self.enums


def pkk_output_header(fh):
    prototypes = [f"const char *\n"
        f"tr_util_{name}_name(enum {name} value);\n" for name in lst_enum_include]

    print(textwrap.dedent("""\
        /*
         * File generated with {program}, please do not edit manually.
         */
        #ifndef {include_header_guard}
        #define {include_header_guard}


        #include "pipe/p_defines.h"
        #include "pipe/p_video_enums.h"


        #ifdef __cplusplus
        extern "C" {{
        #endif

        {prototypes}

        #ifdef __cplusplus
        }}
        #endif

        #endif /* {include_header_guard} */\
        """).format(
            program=pkk_progname,
            include_header_guard=re.sub(r'[^A-Z]', '_', os.path.basename(pkk_cfg.out_header).upper()),
            prototypes="".join(prototypes)
            ), file=fh)


def pkk_output_source(fh):
    print(textwrap.dedent("""\
        /*
         * File generated with {program}, please do not edit manually.
         */
        #include "{include_file}"
        """).format(
            program=pkk_progname,
            include_file=os.path.basename(pkk_cfg.out_header),
            ), file=fh)

    for name in lst_enum_include:
        cases = [f"      case {eid}: return \"{eid}\";\n"
            for eid in enums[name]["data"]]

        print(textwrap.dedent("""\

            const char *
            tr_util_{name}_name(enum {name} value)
            {{
               switch (value) {{
            {cases}
                  default: return "{ucname}_UNKNOWN";
               }}
            }}
            """).format(
                name=name,
                ucname=name.upper(),
                cases="".join(cases)
                ), file=fh)

###
### Main program starts
###
if __name__ == "__main__":
    signal.signal(signal.SIGINT, pkk_signal_handler)

    ### Parse arguments
    pkk_progname = sys.argv[0]
    optparser = pkk_get_argparser()
    pkk_cfg = optparser.parse_args()

    ### Parse input files
    enums = {}
    for file in pkk_cfg.in_files:
        hdrparser = PKKHeaderParser(file)

        try:
            if file != "-":
                with open(file, "r", encoding="UTF-8") as fh:
                    enums.update(hdrparser.parse_file(fh))
            else:
                enums.update(hdrparser.parse_file(sys.stdin))
                break
        except OSError as e:
            pkk_fatal(str(e))

    ### Check if any of the required enums are missing
    errors = False
    for name in lst_enum_include:
        if name not in enums:
            print(f"ERROR: Missing enum '{name}'!")
            errors = True

    if errors:
        pkk_fatal(f"Errors in input. Edit this script ({pkk_progname}) to add/remove included enums.")

    ### Perform output
    if pkk_cfg.out_header:
        with open(pkk_cfg.out_header, "w", encoding="UTF-8") as fh:
            pkk_output_header(fh)

    if pkk_cfg.out_source:
        with open(pkk_cfg.out_source, "w", encoding="UTF-8") as fh:
            pkk_output_source(fh)



#!/usr/bin/env python3

import argparse
import subprocess
import os
import re


def run(args, lo, hi):
    print(f'NIR_SHADER_BISECT_LO: {lo}')
    print(f'NIR_SHADER_BISECT_HI: {hi}')

    env = os.environ.copy()
    env['MESA_SHADER_CACHE_DISABLE'] = '1'
    env['NIR_SHADER_BISECT_LO'] = lo
    env['NIR_SHADER_BISECT_HI'] = hi
    env.pop('MESA_LOG_FILE', None)

    cmd = [args.cmd] + args.cmd_args
    if args.debug:
        print(f"running: {cmd}")
    result = subprocess.run(cmd, env=env,
                            text=True, capture_output=True)
    if args.debug:
        print(f"Result: {result.returncode}")
        print("stdout:")
        print(result.stdout)
        print("stderr:")
        print(result.stderr)

    shaders = set(re.findall(
        "NIR bisect selected source_blake3: (.*) \((.*)\)", result.stderr))
    num = len(shaders)
    print(f'Shaders matched: {num}')
    if num <= 5:
        for blake3, id in sorted(shaders):
            print(f'  {blake3}')
    return shaders


def was_good():
    while True:
        response = input('Was the previous run [g]ood or [b]ad? ')

        if response in ('g', 'b'):
            return response == 'g'


def bisect(args):
    lo = f'{args.lo:064x}'
    hi = f'{args.hi:064x}'

    # Do an initial run to sanity check that the user has actually made
    # nir_shader_bisect_select() select some broken behavior.
    bad = run(args, lo, hi)
    if was_good():
        print("Entire hash range produced a good result -- did you make nir_shader_bisect_select() select the behavior you want?")
        exit(1)

    if len(bad) == 0:
        print("No bad shaders detected -- did you rebuild after adding nir_shader_bisect_select()?")
        exit(1)

    while True:
        if len(bad) == 1:
            for shader in bad:
                print(f"Bisected to source_blake3 {shader}")
                print(
                    f"You can now replace nir_shader_bisect_select() with _mesa_printed_blake3_equal(s->info.source_blake3, (uint32_t[]){shader[0]})")
            exit(0)
        else:
            num = len(bad)
            print(f"Shaders remaining to bisect: {num}")
            if num <= 5:
                for shader in sorted(bad):
                    print(f'  {shader}')

        # Find the middle shader remaining in the set of shaders from the last
        # bad run, and check if the bottom half of set of possibly-bad shaders
        # (up to and including it) is bad.
        ids = sorted([id for blake3, id in bad])
        lo = ids[0]
        split = ids[(len(ids) - 1) // 2]
        cur = run(args, lo, split)

        if was_good():
            bad = bad.difference(cur)
        else:
            bad = cur


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('--lo', type=lambda x: int(x, base=16),
                        default=0,
                        help="NIR_SHADER_BISECT_LO from a bad range, to continue a cancelled run")
    parser.add_argument('--hi', type=lambda x: int(x, base=16),
                        default=(1 << (8 * 32)) - 1,
                        help="NIR_SHADER_BISECT_HI from a bad range, to continue a cancelled run")
    parser.add_argument('cmd')
    parser.add_argument('cmd_args', nargs=argparse.REMAINDER)
    args = parser.parse_args()
    bisect(args)
